
/******************************************************************************

Common routines for mountd and nfsd clients.

Author: Jefferson Ogata (JO317) <ogata@pobox.com>

Copyright 1999 by Jefferson Ogata. All rights reserved.

Please note that this program comes with NO WARRANTY WHATSOEVER. Your use
of this program constitutes your complete acceptance of all liability for
any damage or loss caused by said use.

******************************************************************************/


#include "client.h"


void usage (const int ec, const char *format, ...)
{
    char		**u;
    extern char		*me;
    extern char		*usages[];

    va_list ap;
    va_start (ap, format);
    fprintf (stderr, "%s: ", me);
    vfprintf (stderr, format, ap);
    fprintf (stderr, "\n\n");
    for (u = usages; *u; ++u)
    {
	char		*s;
	int		esc;
	for (s = *u, esc = 0; *s; ++s)
	{
	    if (esc)
	    {
		switch (*s)
		{
		 case 'p':	fputs (me, stderr); break;
		 case '%':	fputc ('%', stderr); break;
		 default:	fputc ('%', stderr); fputc (*s, stderr); break;
		}
		esc = 0;
	    }
	    else
	    {
		switch (*s)
		{
		 case '%':	esc = 1; break;
		 default:	fputc (*s, stderr); break;
		}
	    }
	}
	fputc ('\n', stderr);
    }

    exit (ec);
}


void dumpHandle (const void *fh, const size_t len)
{
    int			k;
    unsigned char	*u = (unsigned char *) fh;

    for (k = 0; k < len; ++k)
	printf ("%02x", u[k]);
}


static int hexValue (char c)
{
    switch (c)
    {
     case '0':			return 0;
     case '1':			return 1;
     case '2':			return 2;
     case '3':			return 3;
     case '4':			return 4;
     case '5':			return 5;
     case '6':			return 6;
     case '7':			return 7;
     case '8':			return 8;
     case '9':			return 9;
     case 'a': case 'A':	return 10;
     case 'b': case 'B':	return 11;
     case 'c': case 'C':	return 12;
     case 'd': case 'D':	return 13;
     case 'e': case 'E':	return 14;
     case 'f': case 'F':	return 15;
    }
    return 0;
}


nfs_fh3 *parseHandle (const char *s)
{
    nfs_fh3		handle;
    unsigned char	*u;
    size_t		len = strlen (s);

    if ((len % 2) || ((len != 2 * NFS_FHSIZE) && (len != 2 * NFS3_FHSIZE)))
	return 0;
    handle.data.data_val = malloc ((handle.data.data_len = len / 2));
    for (u = (unsigned char *) handle.data.data_val; *s; s += 2, ++u)
    {
	int		hi = hexValue (s[0]);
	int		lo = hexValue (s[1]);
	if ((hi < 0) || (lo < 0))
	    return 0;
	*u = (hi << 4) | lo;
    }
    return fhdup (&handle);
}


CLIENT *rpcClient (const u_long program, const u_long version, const int useTcp, char *host, unsigned short *portp, const long localPort)
{
    unsigned short	port = 0;
    char		*s;
    char		*t;
    struct sockaddr_in	sin;
    struct hostent	*h;
    int			sock = RPC_ANYSOCK;
    CLIENT		*client;

    if ((s = strchr (host, ':')))
    {
	port = (unsigned short) strtol (s + 1, &t, 0);
	if (*t || (port == 0))
	    usage (1, "invalid host:port: %s", host);
	*s = '\0';
    }

    /* If local endpoint was specified, try to bind a socket to it. */
    if (localPort >= 0)
    {
	if (useTcp)
	    sock = socket (AF_INET, SOCK_STREAM, 0);
	else
	    sock = socket (AF_INET, SOCK_DGRAM, 0);
	if (sock < 0)
	    usage (2, "cannot create local socket: %s\n", strerror (errno));

	/* Set up local endpoint for binding. */
	memset ((void *) &sin, 0, sizeof (sin));
	sin.sin_family = AF_INET;
	sin.sin_port = htons ((unsigned short) localPort);
	sin.sin_addr.s_addr = htonl (INADDR_ANY);

	/* Bind the local endpoint to the desired address. */
	if (bind (sock, (struct sockaddr *) &sin, sizeof (sin)) < 0)
	    usage (2, "cannot bind local socket to port %d: %s\n", localPort, strerror (errno));
    }

    /* Set up remote endpoint. */
    memset ((void *) &sin, 0, sizeof (sin));
    sin.sin_family = AF_INET;
    sin.sin_port = htons (port);
    if ((h = gethostbyname (host)) == NULL)
	usage (2, "cannot resolve hostname \"%s\"\n", host);
    memcpy ((char *) &sin.sin_addr, h->h_addr, h->h_length);

    if (useTcp)
	client = clnttcp_create (&sin, program, version, &sock, 0, 0);
    else
    {
	struct timeval	tv;

	tv.tv_sec = 3;
	tv.tv_usec = 0;
	client = clntudp_create (&sin, program, version, tv, &sock);
    }

    if (portp)
	*portp = ntohs (sin.sin_port);

    return client;
}


void *memdup (const void *src, const size_t len)
{
    char		*dest = malloc (len);
    memcpy (dest, src, len);
    return dest;
}

#ifdef	NEED_XDR_INT64

/* These routines are not supplied with Red Hat linux. This implementation
   is for i386, and I'm not even sure it's correct. */

bool_t
xdr_uint32(xdrs, objp)
	register XDR *xdrs;
	uint32 *objp;
{
	return xdr_u_int(xdrs, (u_int *) objp);
}

bool_t
xdr_uint64(xdrs, objp)
	register XDR *xdrs;
	uint64 *objp;
{
	uint32 *u = (uint32 *) objp;

#if	BYTE_ORDER == 1234
	if (!xdr_uint32(xdrs, u + 1))
		return (FALSE);
	if (!xdr_uint32(xdrs, u))
		return (FALSE);
#else	/* BYTE_ORDER != 1234 */
#if	BYTE_ORDER == 4321
	if (!xdr_uint32(xdrs, u))
		return (FALSE);
	if (!xdr_uint32(xdrs, u + 1))
		return (FALSE);
#else	/* BYTE_ORDER != 4321 */

Hey man, this byte order is really getting me down.

#endif	/* BYTE_ORDER == 4321 */
#endif	/* BYTE_ORDER == 1234 */
	return (TRUE);
}

#endif	/* !defined(NEED_XDR_INT64) */

