
/******************************************************************************

Simple NFS client

Author: Jefferson Ogata (JO317) <ogata@pobox.com>
Date: 1999/06/07
Version: 0.1

Please note that this program comes with NO WARRANTY WHATSOEVER. Your use
of this program constitutes your complete acceptance of all liability for
any damage or loss caused by said use.

You are free to use or redistribute this program as long as this header
remains intact.

If you have suggestions -- or better yet -- bits of new code, send them
to me and I will deal with them when I have time. The latest version of
this program may always be found at the URL

    http://pobox.com/~ogata/webtools/nfs

commands.c: user commands.

TODO:

Change history:

******************************************************************************/


#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#ifdef	LINUX
#define	__USE_XOPEN
#endif

#include <time.h>
#include <signal.h>
#include <utime.h>
#include <sys/stat.h>
#include <sys/param.h>

#include "client.h"

#include "commands.h"


/* Default argument list for ls, fh commands. */
static char		*defaultArgs[] = { NULL, "." };
#define	NUM_DEFAULT_ARGS	(sizeof(defaultArgs)/sizeof(defaultArgs[0]))


/******************************************************************************
Print a one-line description of the file in the format of ls -alinF to the
specified stream.
******************************************************************************/
static void describe (FILE *stream, nfsObject *f, nfsContext *con)
{
    time_t		t;
    char		*s;
    fattr3		*a = f->attributes;

    fprintf (stream, "%9" PRINTF_INT64_FORMAT "u ", a->fileid);
    switch (a->type)
    {
     case NF3REG:	putc ('-', stream); break;
     case NF3DIR:	putc ('d', stream); break;
     case NF3BLK:	putc ('b', stream); break;
     case NF3CHR:	putc ('c', stream); break;
     case NF3LNK:	putc ('l', stream); break;
     case NF3SOCK:	putc ('S', stream); break;
     case NF3FIFO:	putc ('p', stream); break;
    }
    putc ((a->mode & S_IRUSR) ? 'r' : '-', stream);
    putc ((a->mode & S_IWUSR) ? 'w' : '-', stream);
    if (a->mode & S_IXUSR)
    {
	if (a->mode & S_ISUID)
	    putc ('s', stream);
	else
	    putc ('x', stream);
    }
    else
    {
	if (a->mode & S_ISUID)
	    putc ('S', stream);
	else
	    putc ('-', stream);
    }

    putc ((a->mode & S_IRGRP) ? 'r' : '-', stream);
    putc ((a->mode & S_IWGRP) ? 'w' : '-', stream);
    if (a->mode & S_IXGRP)
    {
	if (a->mode & S_ISGID)
	    putc ('s', stream);
	else
	    putc ('x', stream);
    }
    else
    {
	if (a->mode & S_ISGID)
	    putc ('l', stream);
	else
	    putc ('-', stream);
    }

    putc ((a->mode & S_IROTH) ? 'r' : '-', stream);
    putc ((a->mode & S_IWOTH) ? 'w' : '-', stream);
    if (a->mode & S_IXOTH)
    {
	if (a->mode & S_ISVTX)
	    putc ('t', stream);
	else
	    putc ('x', stream);
    }
    else
    {
	if (a->mode & S_ISVTX)
	    putc ('T', stream);
	else
	    putc ('-', stream);
    }

    fprintf (stream, " %5ld %5ld", (long) a->uid, (long) a->gid);
    fprintf (stream, " %8" PRINTF_INT64_FORMAT "u", a->size);
    t = (time_t) a->mtime.seconds;
    s = ctime (&t);
    if (strchr (s, '\n'))
	*(strchr (s, '\n')) = '\0';
    fprintf (stream, " %s", s);
    fprintf (stream, " %s", f->path);

    if (a->type == NFDIR)
	putc ('/', stream);
    else if (a->type == NFSOCK)
	putc ('=', stream);
    else if (a->type == NFLNK)
    {
	nfspath3	ln = nfsReadlink (f->fh, con);
	if (ln == NULL)
	    putc ('@', stream);
	else
	{
	    fprintf (stream, " -> %s", ln);
	    free (ln);
	}
    }
    else if (a->mode & (S_IXUSR | S_IXGRP | S_IXOTH))
	putc ('*', stream);

    fprintf (stream, "\n");
}


/******************************************************************************
Parse a string into a time_t.
******************************************************************************/
static int parseTime (const char *buf, time_t *time)
{
    char		*s;
    char		**u;
    struct tm		tm;
    static char		*timeFormats[] = {
	"%Y%m%d",
	"%Y%m%d %T",
	"%Y%m%d %R",
	"%Y%m%d %r",
	"%D",
	"%D %T",
	"%D %R",
	"%D %r",
	"%Y/%m/%d",
	"%Y/%m/%d %T",
	"%Y/%m/%d %R",
	"%Y/%m/%d %r",
	NULL
    };

    tzset ();

    for (u = timeFormats; *u; ++u)
    {
	memset (&tm, 0, sizeof (tm));
	s = strptime (buf, *u, &tm);
	if (s && (*s == '\0'))
	    break;
    }

    if (*u == NULL)
	return -1;

    *time = mktime (&tm);

    return 0;
}


/******************************************************************************
Set context atime.
******************************************************************************/
static int atimecmd (size_t argc, char **argv, nfsContext *con)
{
    time_t		t;

    if (argc == 1)
    {
	nfsSetAtime (NULL, con);
	return 0;
    }
    if (argc > 2)
    {
	fprintf (stderr, "%s: too many arguments\n", argv[0]);
	return 0;
    }

    if (parseTime (argv[1], &t) < 0)
    {
	fprintf (stderr, "%s: cannot parse time \"%s\"\n", argv[0], argv[1]);
	return 0;
    }

    nfsSetAtime (&t, con);

    return 0;
}


/******************************************************************************
Set context mtime.
******************************************************************************/
static int mtimecmd (size_t argc, char **argv, nfsContext *con)
{
    time_t		t;

    if (argc == 1)
    {
	nfsSetMtime (NULL, con);
	return 0;
    }
    if (argc > 2)
    {
	fprintf (stderr, "%s: too many arguments\n", argv[0]);
	return 0;
    }

    if (parseTime (argv[1], &t) < 0)
    {
	fprintf (stderr, "%s: cannot parse time \"%s\"\n", argv[0], argv[1]);
	return 0;
    }

    nfsSetMtime (&t, con);

    return 0;
}


/******************************************************************************
Describe a list of files.
******************************************************************************/
static int lscmd (size_t argc, char **argv, nfsContext *con)
{
    size_t		k;
    int			haveFiles = 0;
    nfsDirectory	*files = NULL;
    nfsDirectory	*p;
    nfsObject		*f;

    if (argc == 1)
    {
	defaultArgs[0] = argv[0];
	argv = defaultArgs;
	argc = NUM_DEFAULT_ARGS;
    }

    for (k = 1; k < argc; ++k)
    {
	char		*path = argv[k];

	/* Lookup target. */
	f = nfsFind (path, 0, con);
	if (f == NULL)
	{
	    nfsperror (path);
	    continue;
	}

	/* Create an entry in our file list. */
	p = (nfsDirectory *) malloc (sizeof (*p));
	p->object = f;
	p->next = files;
	files = p;
    }

    /* Sort the list. */
    files = nfsSortDirectory (files);

    /* Do we have any non-directories? If so, we want to list them first
       to mimic the behavior of ls, so we'll check. */
    for (p = files; p; p = p->next)
	if (p->object->attributes->type != NF3DIR)
	    haveFiles = 1;

    /* List non-directories to start with. */
    if (haveFiles)
    {
	for (p = files; p; p = p->next)
	    if (p->object->attributes->type != NF3DIR)
		describe (stdout, p->object, con);
    }

    /* Now list directories. */
    for (p = files, k = 0; p; p = p->next)
    {
	nfsDirectory	*dl;
	nfsDirectory	*d;

	/* Skip non-directories. */
	if (p->object->attributes->type != NF3DIR)
	    continue;

	/* If not first listed directory, or if first directory after
	   list of files, print a blank line. */
	if (k || ((k == 0) && haveFiles))
	    putchar ('\n');

	/* If more than one arg, print name of directory. */
	if (argc > 2)
	    printf ("%s:\n", p->object->path);

	/* It's a directory, so read its contents. */
	dl = nfsReaddir (p->object->fh, con);
	if (dl == NULL)
	    nfsperror ("readdir");

	for (d = dl; d; d = d->next)
	{
	    /* Lookup the file. */
	    f = nfsLookup (p->object->fh, d->object->name, con);
	    if (f == NULL)
	    {
		nfsperror (d->object->name);
		continue;
	    }
	    describe (stdout, f, con);
	    nfsFreeObject (f);
	}

	nfsFreeDirectory (dl);
	++k;
    }

    nfsFreeDirectory (files);

    return 0;
}


/******************************************************************************
Change remote directory.
******************************************************************************/
static int cdcmd (size_t argc, char **argv, nfsContext *con)
{
    char		*path;

    if (argc < 2)
    {
	fprintf (stderr, "%s: not enough arguments\n", argv[0]);
	return 0;
    }
    if (argc > 2)
    {
	fprintf (stderr, "%s: too many arguments\n", argv[0]);
	return 0;
    }

    path = argv[1];
    if (nfsChdir (path, con) < 0)
	nfsperror (path);

    return 0;
}


/******************************************************************************
Change remote root directory.
******************************************************************************/
static int chrootcmd (size_t argc, char **argv, nfsContext *con)
{
    char		*path;

    if (argc < 2)
    {
	fprintf (stderr, "%s: not enough arguments\n", argv[0]);
	return 0;
    }
    if (argc > 2)
    {
	fprintf (stderr, "%s: too many arguments\n", argv[0]);
	return 0;
    }

    path = argv[1];
    if (nfsChroot (path, con) < 0)
	nfsperror (path);

    return 0;
}


/******************************************************************************
Change local directory.
******************************************************************************/
static int lcdcmd (size_t argc, char **argv, nfsContext *con)
{
    char		cwdbuf[MAXPATHLEN + 1];

    if (argc > 2)
    {
	fprintf (stderr, "%s: too many arguments\n", argv[0]);
	return 0;
    }

    if (argc == 2)
	if (chdir (argv[1]) < 0)
	    perror ("chdir");

    if (getcwd (cwdbuf, sizeof (cwdbuf)))
	printf ("Local directory is %s\n", cwdbuf);
    else
	perror ("getcwd");

    return 0;
}


/******************************************************************************
Change credentials.
******************************************************************************/
static int idcmd (size_t argc, char **argv, nfsContext *con)
{
    char		*s;
    char		*t;
    uid_t		uid = nfsGetUser (con);
    gid_t		gid = nfsGetGroup (con);

    if (argc < 2)
    {
	fprintf (stderr, "%s: not enough arguments\n", argv[0]);
	return 0;
    }
    if (argc > 2)
    {
	fprintf (stderr, "%s: too many arguments\n", argv[0]);
	return 0;
    }

    if ((s = strchr (argv[1], '.')))
	*(s++) = '\0';

    if (*argv[1])
    {
	uid = (uid_t) strtol (argv[1], &t, 0);
	if (*t || (uid < 0))
	{
	    fprintf (stderr, "invalid uid \"%s\"\n", argv[1]);
	    return 0;
	}
    }
    if (s)
    {
	gid = (gid_t) strtol (s, &t, 0);
	if (*t || (gid < 0))
	{
	    fprintf (stderr, "invalid gid \"%s\"\n", s);
	    return 0;
	}
    }

    nfsSetUser (uid, con);
    nfsSetGroup (gid, con);

    return 0;
}


/******************************************************************************
Change remote umask.
******************************************************************************/
static int umaskcmd (size_t argc, char **argv, nfsContext *con)
{
    mode_t		umask;
    char		*t;

    if (argc > 2)
    {
	fprintf (stderr, "%s: too many arguments\n", argv[0]);
	return 0;
    }

    if (argc < 2)
    {
	printf ("%03o\n", nfsGetUmask (con));
	return 0;
    }

    umask = (mode_t) strtol (argv[1], &t, 8);
    if (*t || (umask & ~0777))
    {
	fprintf (stderr, "%s: invalid umask\n", argv[1]);
	return 0;
    }

    nfsSetUmask (umask, con);

    return 0;
}


/******************************************************************************
Change modes.
******************************************************************************/
static int chmodcmd (size_t argc, char **argv, nfsContext *con)
{
#define	ALL_PERMS	(S_ISUID | S_ISGID | S_ISVTX | S_IRWXU | S_IRWXG | S_IRWXO)
    mode3		clear = 0;
    mode3		set = 0;
    size_t		k;
    unsigned char	*mode;

    if (argc < 3)
    {
	fprintf (stderr, "%s: not enough arguments\n", argv[0]);
	return 0;
    }

    mode = (unsigned char *) argv[1];
    if (strlen ((char *) mode) == 0)
    {
	fprintf (stderr, "%s: invalid mode\n", argv[0]);
	return 0;
    }

    if (isdigit (mode[0]))
    {
	char		*s;

	set = strtol ((char *) mode, &s, 8);
	if ((*s != '\0') || (set & ~ALL_PERMS))
	{
	    fprintf (stderr, "%s: invalid mode\n", mode);
	    return 0;
	}
	clear = ALL_PERMS;
    }
    else
    {
	unsigned char	*s;
	int		user = 0;
	int		group = 0;
	int		other = 0;
	int		read = 0;
	int		write = 0;
	int		execute = 0;
	int		setuid = 0;
	int		sticky = 0;
	char		op = '\0';

	int		state = 0;

	mode3		bits = 0;

	for (s = mode; *s; ++s)
	{
	    switch (*s)
	    {
	     case 'u':	if (state) break; user = 1; continue;
	     case 'g':	if (state) break; group = 1; continue;
	     case 'o':	if (state) break; other = 1; continue;
	     case 'a':	if (state) break; user = group = other = 1; continue;
	     case '+':	if (state) break; state = 1; op = *s; continue;
	     case '-':	if (state) break; state = 1; op = *s; continue;
	     case '=':	if (state) break; state = 1; op = *s; continue;
	     case 'r':	if (!state) break; read = 1; continue;
	     case 'w':	if (!state) break; write = 1; continue;
	     case 'x':	if (!state) break; execute = 1; continue;
	     case 's':	if (!state) break; setuid = 1; continue;
	     case 't':	if (!state) break; sticky = 1; continue;
	    }
	    fprintf (stderr, "%s: invalid mode\n", mode);
	    return 0;
	}

	if (!(user || group || other))
	    user = group = other = 1;

	if (user && read) bits |= S_IRUSR;
	if (user && write) bits |= S_IWUSR;
	if (user && execute) bits |= S_IXUSR;
	if (group && read) bits |= S_IRGRP;
	if (group && write) bits |= S_IWGRP;
	if (group && execute) bits |= S_IXGRP;
	if (other && read) bits |= S_IROTH;
	if (other && write) bits |= S_IWOTH;
	if (other && execute) bits |= S_IXOTH;
	if (user && setuid) bits |= S_ISUID;
	if (group && setuid) bits |= S_ISGID;
	/* Ignore other && setuid. */
	if (sticky) bits |= S_ISVTX;

	switch (op)
	{
	 case '+':	set = bits; break;
	 case '-':	clear = bits; break;
	 case '=':	set = bits; clear = ALL_PERMS; break;
	}
    }

    for (k = 2; k < argc; ++k)
    {
	nfsObject	*f;

	f = nfsFind (argv[k], 0, con);
	if (f == NULL)
	{
	    nfsperror (argv[k]);
	    continue;
	}
	f->attributes->mode &= ~clear;
	f->attributes->mode |= set;
	if (nfsSetattr (f, NFS_ATTR_MODE, con) < 0)
	    nfsperror (f->path);
	nfsFreeObject (f);
    }

    return 0;
}


/******************************************************************************
Change owner/group.
******************************************************************************/
static int chowncmd (size_t argc, char **argv, nfsContext *con)
{
    char		*s;
    char		*t;
    uid_t		uid = 0;
    gid_t		gid = 0;
    uint32		what = 0;
    size_t		k;

    if (argc < 3)
    {
	fprintf (stderr, "%s: not enough arguments\n", argv[0]);
	return 0;
    }

    if ((s = strchr (argv[1], '.')))
	*(s++) = '\0';

    if (*argv[1])
    {
	uid = (uid_t) strtol (argv[1], &t, 0);
	if (*t || (uid < 0))
	{
	    fprintf (stderr, "invalid uid \"%s\"\n", argv[1]);
	    return 0;
	}
	what |= NFS_ATTR_UID;
    }
    if (s)
    {
	gid = (gid_t) strtol (s, &t, 0);
	if (*t || (gid < 0))
	{
	    fprintf (stderr, "invalid gid \"%s\"\n", s);
	    return 0;
	}
	what |= NFS_ATTR_GID;
    }

    for (k = 2; k < argc; ++k)
    {
	nfsObject	*f;

	f = nfsFind (argv[k], 0, con);
	if (f == NULL)
	{
	    nfsperror (argv[k]);
	    continue;
	}
	if (what & NFS_ATTR_UID)
	    f->attributes->uid = uid;
	if (what & NFS_ATTR_GID)
	    f->attributes->gid = gid;
	if (what && (nfsSetattr (f, what, con) < 0))
	    nfsperror (f->path);
	nfsFreeObject (f);
    }

    return 0;
}


/******************************************************************************
Read filehandles.
******************************************************************************/
static int fhcmd (size_t argc, char **argv, nfsContext *con)
{
    size_t		k;
    nfsDirectory	*files = NULL;
    nfsDirectory	*p;
    nfsObject		*f;

    if (argc == 1)
    {
	defaultArgs[0] = argv[0];
	argv = defaultArgs;
	argc = NUM_DEFAULT_ARGS;
    }

    for (k = 1; k < argc; ++k)
    {
	char		*path = argv[k];

	/* Lookup target. */
	f = nfsFind (path, 0, con);
	if (f == NULL)
	{
	    nfsperror (path);
	    continue;
	}

	/* Create an entry in our file list. */
	p = (nfsDirectory *) malloc (sizeof (*p));
	p->object = f;
	p->next = files;
	files = p;
    }

    /* Sort the list. */
    files = nfsSortDirectory (files);

    /* Now dump the filehandles. */
    for (p = files, k = 0; p; p = p->next)
    {
	dumpHandle (p->object->fh->data.data_val, p->object->fh->data.data_len);
	printf ("   %s\n", p->object->path);
    }

    nfsFreeDirectory (files);

    return 0;
}


/******************************************************************************
Create a link or symbolic link.
******************************************************************************/
static int lncmd (size_t argc, char **argv, nfsContext *con)
{
    int			symbolic = 0;
    char		*target = NULL;
    char		*name = NULL;
    size_t		k;
    nfs_fh3		*dfh;

    for (k = 1; k < argc; ++k)
    {
	if (!strcmp (argv[k], "-s"))
	{
	    symbolic = 1;
	    continue;
	}

	if (strlen (argv[k]) == 0)
	{
	    fprintf (stderr, "%s: invalid argument\n", argv[0]);
	    return 0;
	}

	if (target == NULL)
	{
	    target = argv[k];
	    continue;
	}
	if (name == NULL)
	{
	    name = argv[k];
	    continue;
	}

	fprintf (stderr, "%s: too many arguments\n", argv[0]);
	return 0;
    }

    if (name)
    {
	nfsObject	*f = NULL;

	/* Look up the object. Use parent flag so if the object doesn't
	   exist, but the parent does, we get a useful result. */
	f = nfsFind (name, 1, con);

	if (f == NULL)
	{
	    nfsperror (name);
	    return 0;
	}

	/* Does it exist? */
	if (f->fh == NULL)
	{
	    /* Parent exists. The name will be the last component. */
	    dfh = fhdup (f->dh);
	    name = strdup (f->name);
	    nfsFreeObject (f);
	}
	else
	{
	    /* It does exist. Is it a directory? */
	    if (f->attributes->type == NF3DIR)
	    {
		dfh = fhdup (f->fh);
		name = NULL;
	    }
	    else
	    {
		fprintf (stderr, "%s: %s\n", name, nfserror ((nfsstat3) NFSERR_EXIST));
		return 0;
	    }
	}
    }
    else
    {
	/* No name for link, destination directory is cwd. */
	dfh = nfsGetCwd (con);
    }

    /* If no name specified, or if a directory specified, we still need a
       name. */
    if (name == NULL)
    {
	nfsObject	*f = NULL;

	/* Name will be last component of target. */
	char		*s;
	char		*t = strdup (target);

	/* Cleanup path and find last component. */
	nfsCleanPath (t);
	s = nfsLastComponent (t);

	if ((strlen (s) == 0) || (!strcmp (s, ".")) || (!strcmp (s, "..")))
	{
	    fprintf (stderr, "%s: need name for link\n", argv[0]);
	    free (t);
	    return 0;
	}

	name = strdup (s);
	free (t);

	/* Check if it exists. */
	if ((f = nfsLookup (dfh, name, con)))
	{
	    fprintf (stderr, "%s: %s\n", f->path, nfserror ((nfsstat3) NFSERR_EXIST));
	    nfsFreeObject (f);
	    free (name);
	    return 0;
	}
    }

    /* Make the link. */
    if (symbolic)
    {
	if (nfsSymlink (dfh, name, target, con) < 0)
	    nfsperror (argv[0]);
    }
    else
    {
	nfsObject	*f = NULL;

	/* Find target. */
	f = nfsFind (target, 0, con);
	if (f == NULL)
	    nfsperror (target);
	else
	{
	    if (f->attributes->type == NF3DIR)
		fprintf (stderr, "%s: %s\n", target, nfserror ((nfsstat3) NFSERR_ISDIR));
	    else
	    {
		if (nfsLink (dfh, name, f->fh, con) < 0)
		    nfsperror (argv[0]);
	    }
	    nfsFreeObject (f);
	}
    }

    free (name);
    return 0;
}


/******************************************************************************
Rename a file.
******************************************************************************/
static int mvcmd (size_t argc, char **argv, nfsContext *con)
{
    char		*dest;
    nfsObject		*d = NULL;
    nfs_fh3		*ddir;
    nfsObject		*f;
    int			oneArgument = 1;
    size_t		k;
    int			numargs = 0;
    int			lastarg = 0;
    int			noflag = 0;
    int			nofind = 0;

    /* Make a pass through the arglist to check whether the noflag option is
       set on the last argument. */
    for (k = 1; k < argc; ++k)
    {
	if (!noflag)
	{
	    if (!strcmp (argv[k], "-x"))
	    {
		nofind = !nofind;
		continue;
	    }
	    if (!strcmp (argv[k], "--"))
		continue;
	}

	/* Count actual arguments and keep track of index of the last one. */
	++numargs;
	lastarg = k;
    }

    if (numargs < 2)
    {
	fprintf (stderr, "%s: not enough arguments\n", argv[0]);
	return 0;
    }

    /* Destination is last argument. */
    dest = argv[lastarg];
    if (strlen (dest) == 0)
    {
	fprintf (stderr, "%s: invalid argument\n", argv[0]);
	return 0;
    }

    if (nofind)
    {
	/* Use cwd for destination directory, and literal dest as filename. */
	ddir = nfsGetCwd (con);
    }
    else
    {
	/* Does it exist? */
	d = nfsFind (dest, 1, con);

	if (d == NULL)
	{
	    /* Not even its parent exists. */
	    nfsperror (dest);
	    return 0;
	}

	if (d->fh == NULL)
	{
	    /* Parent exists. The name will be the last component. */
	    ddir = d->dh;
	    dest = d->name;
	}
	else
	{
	    /* It does exist. Is it a directory? */
	    if (d->attributes->type == NF3DIR)
	    {
		oneArgument = 0;
		ddir = d->fh;
	    }
	    else
	    {
		ddir = d->dh;
		dest = d->name;
	    }
	}
    }

    /* If one-argument move, we should have only one argument, dammit. */
    if (oneArgument && (numargs != 2))
    {
	fprintf (stderr, "%s: too many arguments\n", argv[0]);
	return 0;
    }

    nofind = noflag = 0;
    for (k = 1; k < lastarg; ++k)
    {
	if (!noflag)
	{
	    if (!strcmp (argv[k], "-x"))
	    {
		nofind = !nofind;
		continue;
	    }
	    if (!strcmp (argv[k], "--"))
	    {
		noflag = 1;
		continue;
	    }
	}

	/* Find file to be moved. */
	if (nofind)
	    f = nfsLookup (nfsGetCwd (con), argv[k], con);
	else
	    f = nfsFind (argv[k], 0, con);
	if (f == NULL)
	{
	    nfsperror (argv[1]);
	    continue;
	}

	/* If multiple move, use same name for destination. */
	if (!oneArgument)
	    dest = f->name;

	/* Rename it. */
	if (nfsRename (f->dh, f->name, ddir, dest, con) < 0)
	    nfsperror (argv[k]);

	nfsFreeObject (f);
    }

    if (d)
	nfsFreeObject (d);

    return 0;
}


/******************************************************************************
Create directories.
******************************************************************************/
static int mkdircmd (size_t argc, char **argv, nfsContext *con)
{
    size_t		k;

    for (k = 1; k < argc; ++k)
    {
	nfsObject	*f;
	char		*dest = argv[k];

	/* Does it already exist? */
	f = nfsFind (dest, 1, con);

	if (f == NULL)
	{
	    /* Not even parent exists. */
	    nfsperror (argv[k]);
	    continue;
	}

	if (f->fh)
	{
	    /* Object already exists. */
	    fprintf (stderr, "%s: %s\n", argv[k], nfserror (NFS3ERR_EXIST));
	}
	else
	{
	    /* Try to make the directory. */
	    if (nfsMkdir (f->dh, f->name, con) < 0)
		nfsperror (argv[k]);
	}

	nfsFreeObject (f);
    }

    return 0;
}


/******************************************************************************
Remove directories.
******************************************************************************/
static int rmdircmd (size_t argc, char **argv, nfsContext *con)
{
    size_t		k;

    if (argc < 2)
    {
	fprintf (stderr, "%s: not enough arguments\n", argv[0]);
	return 0;
    }

    /* Remove directories. */
    for (k = 1; k < argc; ++k)
    {
	nfsObject	*f;

	f = nfsFind (argv[k], 0, con);
	if (f == NULL)
	{
	    nfsperror (argv[k]);
	    continue;
	}

	if (nfsRmdir (f, con) < 0)
	    nfsperror (f->path);

	nfsFreeObject (f);
    }

    return 0;
}


/******************************************************************************
Remove files.
******************************************************************************/
static int rmcmd (size_t argc, char **argv, nfsContext *con)
{
    size_t		k;

    if (argc < 2)
    {
	fprintf (stderr, "%s: not enough arguments\n", argv[0]);
	return 0;
    }

    /* Remove objects. */
    for (k = 1; k < argc; ++k)
    {
	nfsObject	*f;

	f = nfsFind (argv[k], 0, con);
	if (f == NULL)
	{
	    nfsperror (argv[k]);
	    continue;
	}

	if (nfsRemove (f, con) < 0)
	    nfsperror (f->path);

	nfsFreeObject (f);
    }

    return 0;
}


/******************************************************************************
Create files or update file times.
******************************************************************************/
static int touchcmd (size_t argc, char **argv, nfsContext *con)
{
    size_t		k;

    if (argc < 2)
    {
	fprintf (stderr, "%s: not enough arguments\n", argv[0]);
	return 0;
    }

    for (k = 1; k < argc; ++k)
    {
	nfsObject	*f;
	char		*dest = argv[k];

	/* Does it already exist? */
	f = nfsFind (dest, 1, con);

	if (f == NULL)
	{
	    /* Not even its parent exists. */
	    nfsperror (argv[k]);
	    continue;
	}

	if (f->fh == NULL)
	{
	    /* Try to create the file. */
	    if (nfsCreate (f->dh, f->name, con) < 0)
		nfsperror (argv[k]);
	    else
		f = nfsFind (dest, 0, con);
	}

	if (f && f->fh)
	{
	    /* It already exists, refresh its utimes. */
	    f->attributes->atime.seconds = nfsGetAtime (con);
	    f->attributes->atime.nseconds = 0;
	    f->attributes->mtime.seconds = nfsGetMtime (con);
	    f->attributes->mtime.nseconds = 0;
	    if (nfsSetattr (f, NFS_ATTR_ATIME|NFS_ATTR_MTIME, con) < 0)
	    {
		nfsperror (argv[k]);
		break;
	    }
	}
	else
	{
	}

	nfsFreeObject (f);
    }

    return 0;
}


/******************************************************************************
Pipe file through more.
******************************************************************************/
static int morecmd (size_t argc, char **argv, nfsContext *con)
{
    FILE		*more;
    nfsObject		*f;

    if (argc < 2)
    {
	fprintf (stderr, "%s: not enough arguments\n", argv[0]);
	return 0;
    }
    if (argc > 2)
    {
	fprintf (stderr, "%s: too many arguments\n", argv[0]);
	return 0;
    }

    f = nfsFind (argv[1], 0, con);
    if (f == NULL)
    {
	nfsperror (argv[1]);
	return 0;
    }

    /* Ignore PIPE signals in case user aborts more. */
    signal (SIGPIPE, SIG_IGN);

    /* Send the file contents through more. */
    more = popen ("more", "w");
    if (nfsRead (f, more, con) < 0)
	nfsperror ("read");
    pclose (more);

    /* Restore normal handler for SIGPIPE. */
    signal (SIGPIPE, SIG_DFL);

    return 0;
}


/******************************************************************************
Retrieve file.
******************************************************************************/
static int getcmd (size_t argc, char **argv, nfsContext *con)
{
    FILE		*out;
    size_t		k;
    int64		size;
    struct utimbuf	ut;

    if (argc < 2)
    {
	fprintf (stderr, "%s: not enough arguments\n", argv[0]);
	return 0;
    }

    /* Get the files. */
    for (k = 1; k < argc; ++k)
    {
	nfsObject	*f;

	f = nfsFind (argv[k], 0, con);
	if (f == NULL)
	{
	    nfsperror (argv[k]);
	    continue;
	}

	if (f->attributes->type != NF3REG)
	    fprintf (stderr, "%s: not a regular file\n", f->path);
	else
	{
	    /* Open local file. */
	    out = fopen (f->name, "w");
	    if (out == NULL)
		fprintf (stderr, "%s: %s\n", f->name, strerror (errno));
	    else
	    {
		printf ("%s -> %s: ", f->path, f->name);
		fflush (stdout);

		/* Read NFS file. */
		size = nfsRead (f, out, con);
		if (size < 0)
		{
		    nfsperror (f->path);
		    fclose (out);
		    nfsFreeObject (f);
		    return 0;
		}

		printf (" %" PRINTF_INT64_FORMAT "u bytes\n", size);

		fclose (out);

		/* Set file mode and utimes. */
		ut.modtime = f->attributes->mtime.seconds;
		ut.actime = f->attributes->atime.seconds;
		if (utime (f->name, &ut) < 0)
		    fprintf (stderr, "utime %s: %s\n", f->name, strerror (errno));
		if (chmod (f->name, f->attributes->mode) < 0)
		    fprintf (stderr, "chmod %s: %s\n", f->name, strerror (errno));
	    }
	}

	nfsFreeObject (f);
    }

    return 0;
}


/******************************************************************************
Transmit files.
******************************************************************************/
static int putcmd (size_t argc, char **argv, nfsContext *con)
{
    FILE		*in;
    nfsObject		*f;
    size_t		k;
    int64		size;
    mode_t		umask;
    struct stat		sb;

    if (argc < 2)
    {
	fprintf (stderr, "%s: not enough arguments\n", argv[0]);
	return 0;
    }

    /* Transmit files. */
    for (k = 1; k < argc; ++k)
    {
	char		*path = argv[k];
	char		*name;

	/* Find last component of filename. */
	if ((name = strrchr (path, '/')))
	    ++name;
	else
	    name = path;

	/* Open local file. */
	in = fopen (path, "r");
	if (in == NULL)
	{
	    fprintf (stderr, "%s: %s\n", path, strerror (errno));
	    return 0;
	}

	/* Get file stat. */
	if (fstat (fileno (in), &sb) < 0)
	{
	    perror (path);
	    return 0;
	}

	printf ("%s -> %s: ", path, name);
	fflush (stdout);

	/* Set umask to 077 temporarily for create. */
	umask = nfsGetUmask (con);
	nfsSetUmask (077, con);

	/* Create and find the file. */
	if (nfsCreate (nfsGetCwd (con), name, con) < 0)
	{
	    nfsperror (name);
	    return 0;
	}
	f = nfsFind (name, 0, con);

	/* Restore umask. */
	nfsSetUmask (umask, con);

	/* Write NFS file. */
	size = nfsWrite (f, in, con);
	if (size < 0)
	{
	    nfsperror ("write");
	    fclose (in);
	    nfsFreeObject (f);
	    return 0;
	}

	printf ("%" PRINTF_INT64_FORMAT "u bytes\n", size);

	/* Set file mode and utimes. */
	f->attributes->mode = sb.st_mode;
	f->attributes->mtime.seconds = sb.st_mtime;
	f->attributes->mtime.nseconds = 0;
	f->attributes->atime.seconds = sb.st_atime;
	f->attributes->atime.nseconds = 0;
	if (nfsSetattr (f, NFS_ATTR_MODE|NFS_ATTR_ATIME|NFS_ATTR_MTIME, con) < 0)
	{
	    fprintf (stderr, "setattr %s: %s\n", name, nfserror (nfserrno));
	    return 0;
	}

	/* Attempt to set uid and gid, but don't sweat it. */
	f->attributes->uid = sb.st_uid;
	f->attributes->gid = sb.st_gid;
	nfsSetattr (f, NFS_ATTR_UID|NFS_ATTR_GID, con);

	fclose (in);

	nfsFreeObject (f);
    }

    return 0;
}


/******************************************************************************
Get filesystem static info.
******************************************************************************/
static int ficmd (size_t argc, char **argv, nfsContext *con)
{
    FSINFO3resok	*f;

    if (argc != 1)
    {
	fprintf (stderr, "%s: too many arguments\n", argv[0]);
	return 0;
    }

    f = nfsFsinfo (nfsGetCwd (con), con);
    if (f == NULL)
    {
	nfsperror (argv[0]);
	return 0;
    }

    if (f->obj_attributes.attributes_follow)
    {
	nfsObject	d;
	fattr3		*a;

	a = &f->obj_attributes.post_op_attr_u.attributes;
	d.path = d.name = "";
	d.attributes = a;
	describe (stdout, &d, con);
    }

    printf ("         Read: max %lu, pref %lu, mult %lu\n", f->rtmax, f->rtpref, f->rtmult);
    printf ("        Write: max %lu, pref %lu, mult %lu\n", f->wtmax, f->wtpref, f->wtmult);
    printf ("      Readdir: pref %lu\n", f->dtpref);
    printf ("Max file size: %" PRINTF_INT64_FORMAT "u\n", f->maxfilesize);
    printf ("  Granularity: %ld.%09lus\n", f->time_delta.seconds, f->time_delta.nseconds);
    printf ("   Properties:");
    if (f->properties & FSF3_LINK)
	printf (" links");
    if (f->properties & FSF3_SYMLINK)
	printf (" symlinks");
    if (f->properties & FSF3_HOMOGENEOUS)
	printf (" homogeneous");
    if (f->properties & FSF3_CANSETTIME)
	printf (" settime");
    printf ("\n");

    free (f);

    return 0;
}


/******************************************************************************
Get filesystem dynamic info.
******************************************************************************/
static int fscmd (size_t argc, char **argv, nfsContext *con)
{
    FSSTAT3resok	*f;

    if (argc != 1)
    {
	fprintf (stderr, "%s: too many arguments\n", argv[0]);
	return 0;
    }

    f = nfsFsstat (nfsGetCwd (con), con);
    if (f == NULL)
    {
	nfsperror (argv[0]);
	return 0;
    }

    if (f->obj_attributes.attributes_follow)
    {
	nfsObject	d;
	fattr3		*a;

	a = &f->obj_attributes.post_op_attr_u.attributes;
	d.path = d.name = "";
	d.attributes = a;
	describe (stdout, &d, con);
    }

#define	FS_BYTES_FMT	"%s: %12" PRINTF_INT64_FORMAT "u bytes"
#define	FS_FILES_FMT	", %9" PRINTF_INT64_FORMAT "u files"
    printf (FS_BYTES_FMT, "Total", f->tbytes);
    if (f->tfiles)
	printf (FS_FILES_FMT, f->tfiles);
    printf ("\n");
    printf (FS_BYTES_FMT, " Free", f->fbytes);
    if (f->tfiles)
	printf (FS_FILES_FMT, f->ffiles);
    printf ("\n");
    printf (FS_BYTES_FMT, "Avail", f->abytes);
    if (f->tfiles)
	printf (FS_FILES_FMT, f->afiles);
    printf ("\n");

    printf ("Invar: %12lu s\n", f->invarsec);

    free (f);

    return 0;
}


/******************************************************************************
Exit.
******************************************************************************/
static int exitcmd (size_t argc, char **argv, nfsContext *con)
{
    if (argc != 1)
    {
	fprintf (stderr, "exit: too many arguments\n");
	return 0;
    }
    return 1;
}


/******************************************************************************
******************************************************************************/
struct cmd	commands[] = {
{ "atime",	atimecmd,	ARG_NONE,	"atime time",			"Set atime for object create", },
{ "cd",		cdcmd,		ARG_REMOTE,	"cd dir",			"Change remote working directory", },
{ "chmod",	chmodcmd,	ARG_REMOTE,	"chmod mode obj ...",		"Change permissions on object", },
{ "chown",	chowncmd,	ARG_REMOTE,	"chown [uid].[gid] obj ...",	"Change owner/group on object", },
{ "chroot",	chrootcmd,	ARG_REMOTE,	"chroot dir|##fh",		"Change remote root directory", },
{ "exit",	exitcmd,	ARG_NONE,	"exit",				"Terminate session", },
{ "fh",		fhcmd,		ARG_REMOTE,	"fh file|dir ...",		"Show filehandles", },
{ "fi",		ficmd,		ARG_NONE,	"fi",				"Get static filesystem info", },
{ "fs",		fscmd,		ARG_NONE,	"fs",				"Get dynamic filesystem info", },
{ "get",	getcmd,		ARG_REMOTE,	"get file ...",			"Retrieve files to local directory", },
{ "id",		idcmd,		ARG_NONE,	"id [uid].[gid]",		"Change client credentials", },
{ "lcd",	lcdcmd,		ARG_LOCAL,	"lcd dir",			"Change local working directory", },
{ "ln",		lncmd,		ARG_REMOTE,	"ln [-s] path [link]",		"Make link", },
{ "ls",		lscmd,		ARG_REMOTE,	"ls [file|dir ...]",		"List files", },
{ "mkdir",	mkdircmd,	ARG_REMOTE,	"mkdir dir ...",		"Create directories", },
{ "more",	morecmd,	ARG_REMOTE,	"more file",			"Pipe file through more", },
{ "mtime",	mtimecmd,	ARG_NONE,	"mtime time",			"Set mtime for object create", },
{ "mv",		mvcmd,		ARG_REMOTE,	"mv file1 file2",		"Rename file", },
{ "put",	putcmd,		ARG_LOCAL,	"put file ...",			"Put files to remote directory", },
{ "rm",		rmcmd,		ARG_REMOTE,	"rm file ...",			"Remove files", },
{ "rmdir",	rmdircmd,	ARG_REMOTE,	"rmdir dir ...",		"Remove directories", },
{ "touch",	touchcmd,	ARG_REMOTE,	"touch file ...",		"Update file times or create files", },
{ "umask",	umaskcmd,	ARG_NONE,	"umask [mode]",			"Set remote umask", },
{ NULL,		NULL,		ARG_NONE,	NULL, },
};


