
/******************************************************************************

NFS mountd RPC client

Author: Jefferson Ogata (JO317) <ogata@pobox.com>
Date: 1999/06/07
Version: 0.1

Please note that this program comes with NO WARRANTY WHATSOEVER. Your use
of this program constitutes your complete acceptance of all liability for
any damage or loss caused by said use.

You are free to use or redistribute this program as long as this header
remains intact.

If you have suggestions -- or better yet -- bits of new code, send them
to me and I will deal with them when I have time. The latest version of
this program may always be found at the URL

    http://pobox.com/~ogata/webtools/nfs

This program uses the RPC mountd protocol to enumerate the exported mount
points on a remote host. It accepts a host and optional mountd port number
on the command line, and prints out any mount points, client lists, and
filehandles it can retrieve from the specified host.

Change history:

******************************************************************************/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "client.h"

#include "mount.h"


char *me = 0;

char *usages[] = {
"usage: %p [-3tn] [-p port] [-u uid] [-g gid] [-h host] host[:port] [path ...]",
"",
"Uses the RPC mountd protocol to enumerate the list of exported mount points",
"on a remote host. The :port option may be used to specify the port on which",
"to contact the remote system\'s mountd; if omitted, the portmapper on the",
"remote system will be consulted. For each exported mount point, the program",
"calls the mount RPC, reports the filehandle returned by mountd, and calls",
"the umount RPC. Note that this does not actually mount the filesystem; it",
"merely queries mountd for the mount filehandle. If a list of paths is given",
"on the command line, %p queries only for the handles of those paths, and",
"does not enumerate the mount points. If a mount request is refused by the",
"remote mountd, the error message may indicate whether the target exists.",
"It is a common vulnerability in mountd implementations that they leak the",
"existence of files to unauthorized hosts.",
"",
"Other options are:",
"-3        Use version 3 of the mount protocol. The default is version 1.",
"          Note that filehandles returned from version 3 may be NFS3",
"          handles, and may be invalid for version 2 NFS access.",
"-t        Use TCP for the RPC connection. The default is to use UDP.",
"-n        Do not call the mount RPC, but enumerate the exported mount",
"          only.",
"-p port   Set the port number of the local endpoint for the RPC connection.",
"          Some mountd implementations will only honor requests from a",
"          privileged port. This option does not work with -t on some",
"          platforms.",
"If any of the following options is used, UNIX RPC authentication is",
"enabled. If none is used, no RPC authentication is used, and the defaults",
"are irrelevant:",
"-u uid    Set the uid used in RPC authentication. The default is the uid",
"          of the invoking user.",
"-g gid    Set the gid used in RPC authentication. The default is the gid",
"          of the invoking user.",
"-h host   Set the hostname used in RPC authentication. The default is",
"          \"localhost\".",
NULL
};


static char *mnterror (u_int status)
{
    switch (status)
    {
     case MNT_OK:		return "No error";
     case MNT3ERR_PERM:		return "Not owner";
     case MNT3ERR_NOENT:	return "No such file or directory";
     case MNT3ERR_IO:		return "I/O error";
     case MNT3ERR_ACCES:	return "Permission denied";
     case MNT3ERR_NOTDIR:	return "Not a directory";
     case MNT3ERR_INVAL:	return "Invalid argument";
     case MNT3ERR_NAMETOOLONG:	return "File name too long";
     case MNT3ERR_NOTSUPP:	return "Operation not supported";
     case MNT3ERR_SERVERFAULT:	return "Server fault";
     default:			return strerror ((int) status);
    }
}


/* Try to mount a remote path, and report any filehandle we get. */
static void tryMount (char *path, u_long version, CLIENT *client)
{
    void		*fh;
    size_t		fl;

    /* Try to mount it so we can get a filehandle. */
    if (version == MOUNTVERS)
    {
	fhstatus		*fs;

	/* Call the mount RPC. */
	fs = mountproc_mnt_1 (&path, client);

	if (fs == NULL)
	{
	    printf ("    mount failed\n");
	    return;
	}
	else if (fs->fhs_status != MNT_OK)
	{
	    printf ("    %s\n", mnterror (fs->fhs_status));
	    return;
	}

	/* Get the handle. */
	fl = sizeof (fs->fhs_fh);
	fh = malloc (fl);
	memcpy (fh, &fs->fhs_fh, fl);

	/* Unmount the filesystem. */
	mountproc_umnt_1 (&path, client);
    }
    else
    {
	mountres3		*mr;
	fhandle3		*fs;

	/* Call the mount RPC. */
	mr = mountproc_mnt_3 (&path, client);

	if (mr == NULL)
	{
	    printf ("    mount failed\n");
	    return;
	}
	else if (mr->fhs_status != MNT_OK)
	{
	    printf ("    %s\n", mnterror (mr->fhs_status));
	    return;
	}

	/* Copy the filehandle. */
	fs = &mr->mountres3_u.mountinfo.fhandle;
	fl = fs->fhandle3_len;
	fh = malloc (fl);
	memcpy (fh, fs->fhandle3_val, fl);

	/* Unmount the filesystem. */
	mountproc_umnt_3 (&path, client);
    }

    /* Print out the filehandle we got. */
    printf ("    ");
    dumpHandle (fh, fl);
    printf ("\n");

    free (fh);
}


int main (int argc, char **argv)
{
    char		*host = NULL;
    char		*clientHost = NULL;
    long		uid = -1;
    long		gid = -1;
    int			useTcp = 0;
    unsigned short	port;
    long		localPort = -1;
    int			getHandles = 1;
    u_long		version = MOUNTVERS;
    char		**paths;
    char		**nextPath;

    int			k;
    CLIENT		*client;

    if ((me = strrchr (argv[0], '/')))
	++me;
    else
	me = argv[0];

    nextPath = paths = (char **) malloc ((argc + 1) * sizeof (paths[0]));

    for (k = 1; k < argc; ++k)
    {

	if (argv[k][0] == '-')
	{
	    int			j;
	    char		*s;
	    char		*t;

	    for (s = argv[k] + 1, j = k; *s; ++s)
	    {
		switch (*s)
		{
		 case 'h':
		    if ((++j) >= argc)
			usage (1, "-%c requires an argument", *s);
		    clientHost = argv[j];
		    continue;
		 case 'u':
		    if ((++j) >= argc)
			usage (1, "-%c requires an argument", *s);
		    uid = strtol (argv[j], &t, 0);
		    if (*t || (uid < 0))
			usage (1, "invalid uid argument: %s", argv[j]);
		    continue;
		 case 'g':
		    if ((++j) >= argc)
			usage (1, "-%c requires an argument", *s);
		    gid = strtol (argv[j], &t, 0);
		    if (*t || (gid < 0))
			usage (1, "invalid uid argument: %s", argv[j]);
		    continue;
		 case 'p':
		    if ((++j) >= argc)
			usage (1, "-%c requires an argument", *s);
		    localPort = strtol (argv[j], &t, 0);
		    if (*t || (localPort < 0))
			usage (1, "invalid port argument: %s", argv[j]);
		    continue;
		 case 'n':
		    getHandles = 0;
		    continue;
		 case 't':
		    useTcp = 1;
		    continue;
		 case '3':
		    version = MOUNTVERS3;
		    continue;

		 default:
		    usage (1, "unknown option: -%c", *s);
		}
	    }

	    k = j;

	    continue;
	}
	if (host == NULL)
	{
	    host = argv[k];
	    continue;
	}

	*nextPath = argv[k];
	++nextPath;
    }

    *nextPath = 0;

    if (host == NULL)
	usage (1, "not enough arguments");

    /* Create RPC client. */
    client = rpcClient (MOUNTPROG, version, useTcp, host, &port, localPort);
    if (client == NULL)
    {
	clnt_pcreateerror ("clnt_create");
	return errno ? errno : 1;
    }

    fprintf (stderr, "Connected to mountd server at %s %s:%d.\n", useTcp ? "tcp" : "udp", host, port);

    /* Create RPC authentication. */
    if (client->cl_auth)
	auth_destroy (client->cl_auth);
    if (clientHost || (uid >= 0) || (gid >= 0))
    {
	uid_t		u;
	gid_t		g;

	if (uid < 0)
	    uid = getuid ();
	if (gid < 0)
	    gid = getgid ();
	if (clientHost == NULL)
	    clientHost = "localhost";

	u = (uid_t) uid;
	g = (gid_t) gid;
	client->cl_auth = authunix_create (clientHost, u, g, 1, &g);
    }
    else
    {
	client->cl_auth = authnone_create ();
    }

    if (*paths)
    {
	char		**u;

	/* Walk the path list. */
	for (u = paths; *u; ++u)
	{
	    printf ("%s:\n", *u);
	    if (getHandles)
		tryMount (*u, version, client);
	}
    }
    else
    {
	exports		*ep;
	exports		e;
	groups		g;

	/* Get the export list. */
	if (version == MOUNTVERS)
	    ep = mountproc_export_1 (NULL, client);
	else
	    ep = mountproc_export_3 (NULL, client);

	if (ep == NULL)
	{
	    fprintf (stderr, "No exports from %s:%d\n", host, port);
	    return 1;
	}

	/* Walk the export list. */
	for (e = *ep; e; e = e->ex_next)
	{
	    /* Describe the export. */
	    printf ("%s:", e->ex_dir);
	    for (g = e->ex_groups; g; g = g->gr_next)
		printf (" %s", g->gr_name);
	    printf ("\n");

	    /* Try to get a filehandle. */
	    if (getHandles)
		tryMount (e->ex_dir, version, client);
	}
    }

    return 0;
}


