
/******************************************************************************

Simple NFS client

Author: Jefferson Ogata (JO317) <ogata@pobox.com>
Date: 1999/06/07
Version: 0.1

Please note that this program comes with NO WARRANTY WHATSOEVER. Your use
of this program constitutes your complete acceptance of all liability for
any damage or loss caused by said use.

You are free to use or redistribute this program as long as this header
remains intact.

If you have suggestions -- or better yet -- bits of new code, send them
to me and I will deal with them when I have time. The latest version of
this program may always be found at the URL

    http://pobox.com/~ogata/webtools/nfs

This program is a simple NFS client. It accepts a filehandle on the command
line, along with a host and optional NFS port number. It will explore and
retrieve files exported under the given filehandle from the specified host.

Change history:

******************************************************************************/


#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/param.h>

#ifdef	USE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif

#include "client.h"

#include "commands.h"
#include "nfsprocs.h"
#include "nfs_prot.h"


/* Who am I? */
char *me = 0;


char *usages[] = {
"usage: %p [-3t] [-p port] host[:port] fh",
"",
"%p is a simple FTP-like NFS client program. It accepts a host, optional",
"port number, and filehandle on the command line, and provides a shell with",
"basic file management commands such as ls, mv, etc. A complete list of",
"commands can be viewed by typing a ? at the command prompt.",
"",
"The prompt informs you what your path is on the remote NFS filesystem",
"relative to the initial filehandle you provided, and what uid and gid you",
"appear as to the remote server. The uid and gid may be changed with the id",
"command. Note that uid and gid 0 are commonly remapped by servers to",
"nobody and nogroup respectively.",
"",
"Other options are:",
"-3        Use version 3 of the NFS protocol. The default is version 2.",
"          Note that you may not be able to use a version 2 handle with",
"          NFS version 3.",
"-t        Use TCP for the RPC connection. The default is to use UDP.",
"-p port   Set the port number of the local endpoint for the RPC connection.",
"          Some NFS implementations will only honor requests from a",
"          privileged port. This option does not work with -t on some",
"          platforms.",
"",
"If readline support is compiled in "
#ifdef	USE_READLINE
"(it is)"
#else
"(it isn\'t)"
#endif
" command and filename completion",
"are available.",
NULL
};


/* The context in which we live. We need to get to this from the readline
   callbacks, so it has to be out here in the open. */
static nfsContext	*context;


static char **parseCommand (char *buf, size_t *argcp)
{
#define	MAXARGS	1024
    static unsigned char	*argv[MAXARGS];
    unsigned char		*s;
    unsigned char		*t;
    unsigned char		**u;

    /* Skip leading space. */
    for (s = (unsigned char *) buf; isspace (*s); ++s)
	;

    /* If there's anything, save this line on the history. */
#ifdef	USE_READLINE
    if (*s)
	add_history (buf);
#endif

    /* Break into words. */
    for (u = argv; u < (argv + MAXARGS) && *s; ++u)
    {
	char		quote;
	int		esc;
	unsigned char	c;
	for (*u = t = s, quote = '\0', esc = 0; (c = *s); ++s)
	{
	    if (esc)
	    {
		*(t++) = c;
		esc = 0;
		continue;
	    }
	    if (c == '\\')
	    {
		esc = 1;
		continue;
	    }
	    if (quote == '\0')
	    {
		/* Not in any quoted word. */
		if (isspace (c))
		{
		    /* End of this word. */
		    break;
		}
		if ((c == '\'') || (c == '\"'))
		{
		    quote = c;
		    continue;
		}
	    }
	    else if (c == quote)
	    {
		quote = '\0';
		continue;
	    }
	    *(t++) = c;
	}

	/* Skip trailing space. */
	for ( ; isspace (*s); ++s)
	    ;

	/* End last word. */
	*t = '\0';
    }
    *argcp = u - argv;

    return (char **) argv;
}


static char *readCommand (char *prompt)
{
#ifdef	USE_READLINE
    static char		*buf = NULL;
#else
    static char		buf[NFS_MAXPATHLEN * 3];
#endif

    /* Read a command. */
#ifdef	USE_READLINE
    if (buf)
	free (buf);
    if ((buf = readline (prompt)) == NULL)
	return NULL;
#else
    fputs (prompt, stdout);
    fflush (stdout);
    if ((fgets (buf, sizeof (buf), stdin)) == NULL)
	return NULL;
#endif

    return buf;
}


#ifdef	USE_READLINE
static char *commandGenerator (char *text, int state)
{
    static int		cursor = 0;
    static size_t	len = 0;
    char		*name;

    if (state == 0)
    {
	cursor = 0;
	len = strlen (text);
    }

    /* Look for a matching name. */
    while ((name = commands[cursor].cmd))
    {
	++cursor;
	if (!strncmp (name, text, len))
	    return strdup (name);
    }

    /* If no names matched, return NULL. */
    return NULL;
}


static char *remoteGenerator (char *text, int state)
{
    nfsDirectory		*next;
    static nfsDirectory		*dl = NULL;
    static char			*hunt = NULL;
    static size_t		len = 0;
    static size_t		dirlen = 0;
    static nfs_fh3		*fh = NULL;

    if (state == 0)
    {
	char			*path = strdup (text);
	char			*dir = ".";
	nfsObject		*f;

	/* If a directory has been typed, use it. */
	if ((hunt = strrchr (path, '/')))
	{
	    *hunt = '\0';
	    ++hunt;
	    dir = path;
	}
	else
	    hunt = path;
	if (*dir == '\0')
	    dir = "/";

	/* Look up the directory. */
	f = nfsFind (dir, 0, context);
	if (f == NULL)
	    return NULL;

	/* Save directory handle. */
	if (fh)
	    fhfree (fh);
	fh = fhdup (f->fh);

	/* Read its contents. */
	if (dl)
	    nfsFreeDirectory (dl);
	dl = nfsReaddir (f->fh, context);
	if (dl == NULL)
	    nfsperror ("readdir");

	/* How much of it is the directory? */
	dirlen = hunt - path;

	/* We only want to match against the last component. */
	hunt = text + dirlen;
	len = strlen (hunt);

	nfsFreeObject (f);
    }

    /* Find next node with a name that matches. */
    for ( ; dl; )
    {
	char		*path;
	char		*name;

	next = dl->next;

	/* Skip . and .. */
	if ((!strcmp (dl->object->name, ".")) || (!strcmp (dl->object->name, "..")))
	{
	    nfsFreeDirectoryNode (dl);
	    dl = next;
	    continue;
	}

	path = malloc (dirlen + strlen (dl->object->name) + 2);
	strncpy (path, text, dirlen);
	name = path + dirlen;
	strcpy (name, dl->object->name);
	nfsFreeDirectoryNode (dl);
	dl = next;

	if (!strncmp (hunt, name, len))
	{
	    nfsObject	*f;

	    /* It matches, find out more about it. */
	    f = nfsLookup (fh, name, context);
	    if (f == NULL)
	    {
		free (path);
		continue;
	    }

	    /* If it's a directory, tack on a /. */
	    if (f->attributes->type == NF3DIR)
		strcat (name, "/");

	    nfsFreeObject (f);

	    return path;
	}
	
	free (path);
    }

    return NULL;
}


static char **completionFunction (char *text, int start, int end)
{
    char		**matches = NULL;
    struct cmd		*cmd;

    /* If completing first word, use command list. */
    if (start == 0)
    {
	matches = completion_matches (text, commandGenerator);
    }
    else
    {
	char			*buf;
	char			**argv;
	size_t			argc;
	enum argtype		at = ARG_NONE;

	/* Copy first part of string. */
	buf = malloc (start + 1);
	strncpy (buf, rl_line_buffer, start);
	buf[start] = '\0';

	/* Parse it. */
	argv = parseCommand (buf, &argc);

	/* Look for the command. */
	if (argc > 0)
	{
	    for (cmd = commands; cmd->cmd; ++cmd)
		if (!strcmp (argv[0], cmd->cmd))
		{
		    at = cmd->argType;
		    break;
		}
	}

	/* Now try to match an arg of the appropriate type. */
	switch (at)
	{
	 case ARG_LOCAL:
	    matches = completion_matches (text, filename_completion_function);
	    break;

	 case ARG_REMOTE:
	    matches = completion_matches (text, remoteGenerator);
	    if (matches && matches[0] && (matches[1] == NULL))
	    {
		size_t		len = strlen (matches[0]);

		/* Only one completion. If it's a directory, don't append. */
		if (len && (matches[0][len - 1] == '/'))
		    rl_completion_append_character = '\0';
		else
		    rl_completion_append_character = ' ';
	    }
	    else
		rl_completion_append_character = ' ';
	    break;

	 case ARG_NONE:
	    break;
	}

	free (buf);
    }

    rl_attempted_completion_over = 1;
    return matches;
}
#endif	/* defined(USE_READLINE) */


/******************************************************************************
******************************************************************************/
static void interpreter (nfs_fh3 *root, CLIENT *client, const u_long version)
{
    struct cmd		*cmd;

    /* Create new NFS context. */
    context = nfsNewContext (root, client, version);

#ifdef	USE_READLINE
    /* Initialize readline history mechanism. */
    using_history ();

    /* Tell readline that we want a crack first. */
    rl_attempted_completion_function = completionFunction;
#endif

    while (1)
    {
	char		*buf;
	char		**argv;
	size_t		argc;
	char		*path;
	unsigned char	*s;
	static char	*prompt = NULL;
	static size_t	promptLength = 0;

	/* Create command prompt. */
	path = nfsGetPath (context);

#define	CREDS_SIZE	64
	if (strlen (path) + CREDS_SIZE > promptLength)
	{
	    if (prompt)
		free (prompt);
	    promptLength = CREDS_SIZE + ((strlen (path) + MAXPATHLEN - 1) / MAXPATHLEN) * MAXPATHLEN;
	    prompt = malloc (promptLength);
	}

	sprintf (prompt, "%s %ld.%ld> ", path, (long) nfsGetUser (context), (long) nfsGetGroup (context));

	/* Get command line. */
	buf = readCommand (prompt);

	/* Quit on EOF. */
	if (buf == NULL)
	{
	    putchar ('\n');
	    return;
	}

	/* Check for bang-escape. */
	for (s = (unsigned char *) buf; *s && isspace (*s); ++s)
	    ;
	if (*s == '!')
	{
	    for (++s; *s && isspace (*s); ++s)
		;
	    if (*s)
		system ((char *) s);
#ifdef	USE_READLINE
	    add_history (buf);
#endif
	    continue;
	}

	/* Parse into words. */
	argv = parseCommand (buf, &argc);

	if (argc == 0)
	    continue;

	if (!strcmp (argv[0], "?"))
	{
	    size_t	maxLength = 0;

	    for (cmd = commands; cmd->cmd; ++cmd)
		if (strlen (cmd->usage) > maxLength)
		    maxLength = strlen (cmd->usage);
	    for (cmd = commands; cmd->cmd; ++cmd)
		if ((argc == 1) || !strcmp (argv[1], cmd->cmd))
		    fprintf (stderr, "%-*s  %s\n", maxLength, cmd->usage, cmd->description);
	    if ((argc == 1) || !strcmp (argv[1], "!"))
		fprintf (stderr, "%-*s  %s\n", maxLength, "!cmd", "Execute cmd in a subshell\n");
	    continue;
	}

	for (cmd = commands; cmd->cmd; ++cmd)
	    if (!strcmp (argv[0], cmd->cmd))
		break;

	if (cmd->cmd == NULL)
	{
	    fprintf (stderr, "%s: command not found\n", argv[0]);
	    continue;
	}

	if ((*cmd->fn) (argc, argv, context))
	    return;
    }
}


int main (int argc, char **argv)
{
    char		*host = NULL;
    unsigned short	port = 0;
    int			useTcp = 0;
    u_long		version = NFS_VERSION;
    long		localPort = -1;
    char		*mountPoint = NULL;

    int			k;
    CLIENT		*client;
    nfs_fh3		handle;
    nfs_fh3		*fh = NULL;

    if ((me = strrchr (argv[0], '/')))
	++me;
    else
	me = argv[0];

    for (k = 1; k < argc; ++k)
    {
	if (argv[k][0] == '-')
	{
	    int			j;
	    char		*s;
	    char		*t;

	    for (s = argv[k] + 1, j = k; *s; ++s)
	    {
		switch (*s)
		{
		 case 't':
		    useTcp = 1;
		    continue;

		 case 'p':
		    if ((++j) >= argc)
			usage (1, "-%c requires an argument", *s);
		    localPort = strtol (argv[j], &t, 0);
		    if (*t)
			usage (1, "invalid port argument: %s", argv[j]);
		    continue;

		 case '3':
		    version = NFS_V3;
		    continue;

		 case 'm':
		    if ((++j) >= argc)
			usage (1, "-%c requires an argument", *s);
		    mountPoint = argv[j];
		    continue;

		 default:
		    usage (1, "unknown option: -%c", *s);
		}
	    }

	    k = j;

	    continue;
	}
	if (host == NULL)
	{
	    host = argv[k];
	    continue;
	}
	if (fh == NULL)
	{
	    unsigned char	*u;
	    size_t		len = strlen (argv[k]);

	    if ((len % 2) || (len == 0) || (len > 2 * NFS3_FHSIZE))
		usage (1, "handle has invalid length");
	    handle.data.data_len = len / 2;
	    handle.data.data_val = malloc (len);
	    u = (unsigned char *) handle.data.data_val;
	    if ((fh = parseHandle (argv[k])) == 0)
		usage (1, "invalid file handle");
	    continue;
	}
	usage (1, "too many arguments");
    }

    if (fh == NULL)
	usage (1, "not enough arguments");

    /* Handle a mount request. */
    if (mountPoint)
    {
    }

    client = rpcClient (NFS_PROGRAM, version, useTcp, host, &port, localPort);

    if (client == NULL)
    {
	clnt_pcreateerror ("clnt_create");
	return errno ? errno : 1;
    }

    printf ("Connected to nfs server at %s %s:%d.\n", useTcp ? "tcp" : "udp", host, port);

    interpreter (fh, client, version);

    return 0;
}

