
/******************************************************************************

Simple NFS client

Author: Jefferson Ogata (JO317) <ogata@pobox.com>
Date: 1999/06/07
Version: 0.1

Please note that this program comes with NO WARRANTY WHATSOEVER. Your use
of this program constitutes your complete acceptance of all liability for
any damage or loss caused by said use.

You are free to use or redistribute this program as long as this header
remains intact.

If you have suggestions -- or better yet -- bits of new code, send them
to me and I will deal with them when I have time. The latest version of
this program may always be found at the URL

    http://pobox.com/~ogata/webtools/nfs

nfsprocs.c: version-independent wrappers for NFS RPCs.

TODO:
pathconf

Change history:

******************************************************************************/


#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "client.h"

#include "nfsprocs.h"


/******************************************************************************
The current error status.
******************************************************************************/
nfsstat3	nfserrno = (nfsstat3) NFS_OK;


/******************************************************************************
Generate an error message for an nfsstat3.
******************************************************************************/
char *nfserror (const nfsstat3 status)
{
    switch (status)
    {
     case NFS_OK:		return "No error";
     case NFSERR_PERM:		return "Not owner";
     case NFSERR_NOENT:		return "No such file or directory";
     case NFSERR_IO:		return "I/O error";
     case NFSERR_NXIO:		return "No such device or address";
     case NFSERR_ACCES:		return "Permission denied";
     case NFSERR_EXIST:		return "File exists";
     case NFSERR_XDEV:		return "Cross-device link";
     case NFSERR_NODEV:		return "No such device";
     case NFSERR_NOTDIR:	return "Not a directory";
     case NFSERR_ISDIR:		return "Is a directory";
     case NFSERR_INVAL:		return "Invalid argument";
     case NFSERR_FBIG:		return "File too large";
     case NFSERR_NOSPC:		return "No space left on device";
     case NFSERR_ROFS:		return "Read-only file system";
     case NFSERR_OPNOTSUPP:	return "Operation not supported";
     case NFSERR_NAMETOOLONG:	return "File name too long";
     case NFSERR_NOTEMPTY:	return "Directory not empty";
     case NFSERR_DQUOT:		return "Disk quota exceeded";
     case NFSERR_STALE:		return "Stale NFS file handle";
     case NFSERR_REMOTE:	return "Object is remote";
     case NFSERR_WFLUSH:	return "Write cache flushed";
     case NFS3ERR_BADHANDLE:	return "NFS3 bad handle";
     case NFS3ERR_NOT_SYNC:	return "Object out of sync";
     case NFS3ERR_BAD_COOKIE:	return "NFS3 bad cookie";
     case NFS3ERR_NOTSUPP:	return "NFS3 operation not supported";
     case NFS3ERR_TOOSMALL:	return "NFS3 object too small";
     case NFS3ERR_SERVERFAULT:	return "NFS3 server fault";
     case NFS3ERR_BADTYPE:	return "NFS3 bad type";
     case NFS3ERR_JUKEBOX:	return "NFS3 jukebox error";
     default:			return strerror ((int) status);
    }
}


/******************************************************************************
Print an error message a la perror for the current value of nfserrno.
******************************************************************************/
void nfsperror (const char *msg)
{
    fprintf (stderr, "%s: %s\n", msg, nfserror (nfserrno));
}


/******************************************************************************
Free a filehandle created by fhdup or fh2dup.
******************************************************************************/
void fhfree (nfs_fh3 *fh)
{
    free (fh->data.data_val);
    free (fh);
}


/******************************************************************************
Duplicate a filehandle. Returns malloced memory. The caller is responsible
for freeing the result by calling fhfree.
******************************************************************************/
nfs_fh3 *fhdup (const nfs_fh3 *src)
{
    nfs_fh3		*new;
    new = (nfs_fh3 *) malloc (sizeof (*new));
    new->data.data_len = src->data.data_len;
    new->data.data_val = memdup (src->data.data_val, new->data.data_len);
    return new;
}


/******************************************************************************
Duplicate a version 2 filehandle. Returns malloced memory. The caller is
responsible for freeing the result by calling fhfree.
******************************************************************************/
nfs_fh3 *fh2dup (const nfs_fh *src)
{
    nfs_fh3		*new;
    new = (nfs_fh3 *) malloc (sizeof (*new));
    new->data.data_len = NFS_FHSIZE;
    new->data.data_val = memdup (src->data, new->data.data_len);
    return new;
}


/******************************************************************************
Given a version 3 filehandle, returns a pointer to a version 2 filehandle.
This is simply the value of the handle. If the handle has the wrong size
for a version 2 handle, aborts the program.
******************************************************************************/
nfs_fh *fh3to2 (const nfs_fh3 *fh)
{
    if (fh->data.data_len != NFS_FHSIZE)
    {
	fprintf (stderr, "attempt to coerce non-v2 handle to v2 handle!\n");
	exit (3);
    }
    return (nfs_fh *) fh->data.data_val;
}


/******************************************************************************
Return -1, 0, or 1 as filehandle a is less than, equal to, or greater than
filehandle b.
******************************************************************************/
int fhcmp (const nfs_fh3 *a, const nfs_fh3 *b)
{
    int			cmp;

    if (a == b)
	return 0;

    if (a->data.data_len < b->data.data_len)
    {
	cmp = memcmp (a->data.data_val, b->data.data_val, a->data.data_len);
	if (cmp) return cmp;
	return -1;
    }

    if (b->data.data_len < a->data.data_len)
    {
	cmp = memcmp (a->data.data_val, b->data.data_val, b->data.data_len);
	if (cmp) return cmp;
	return 1;
    }

    return memcmp (a->data.data_val, b->data.data_val, a->data.data_len);
}


/******************************************************************************
Translate the contents of a version 2 fattr to a version 3 fattr3.
******************************************************************************/
static void attr2to3 (fattr3 *a3, const fattr *a2)
{
    switch (a2->type)
    {
     case NFREG:	a3->type = NF3REG; break;
     case NFDIR:	a3->type = NF3DIR; break;
     case NFBLK:	a3->type = NF3BLK; break;
     case NFCHR:	a3->type = NF3CHR; break;
     case NFLNK:	a3->type = NF3LNK; break;
     case NFSOCK:	a3->type = NF3SOCK; break;
     case NFFIFO:	a3->type = NF3FIFO; break;
     default:
	fprintf (stderr, "unsupported v2 filetype %d\n", a2->type);
	a3->type = NF3REG;	/* What the hell... lie about it. */
	break;
    }
    a3->mode = a2->mode;
    a3->nlink = a2->nlink;
    a3->uid = a2->uid;
    a3->gid = a2->gid;
    a3->size = a2->size;
    a3->used = 0;
    a3->rdev.specdata1 = a2->rdev;
    a3->rdev.specdata2 = 0;
    a3->fsid = a2->fsid;
    a3->fileid = a2->fileid & 0xfffffff;		/* mask top bits */
    a3->atime.seconds = a2->atime.seconds;
    a3->atime.nseconds = a2->atime.useconds * 1000;	/* usec to nsec */
    a3->mtime.seconds = a2->mtime.seconds;
    a3->mtime.nseconds = a2->mtime.useconds * 1000;
    a3->ctime.seconds = a2->ctime.seconds;
    a3->ctime.nseconds = a2->ctime.useconds * 1000;
}


/******************************************************************************
Translate the contents of a version 3 fattr3 to a version 2 fattr.
******************************************************************************/
static void attr3to2 (fattr *a2, const fattr3 *a3)
{
    switch (a3->type)
    {
     case NF3REG:	a2->type = NFREG; break;
     case NF3DIR:	a2->type = NFDIR; break;
     case NF3BLK:	a2->type = NFBLK; break;
     case NF3CHR:	a2->type = NFCHR; break;
     case NF3LNK:	a2->type = NFLNK; break;
     case NF3SOCK:	a2->type = NFSOCK; break;
     case NF3FIFO:	a2->type = NFFIFO; break;
     default:
	fprintf (stderr, "unsupported v3 filetype %d\n", a3->type);
	a2->type = NFREG;
	break;
    }
    a2->mode = a3->mode;
    a2->nlink = a3->nlink;
    a2->uid = a3->uid;
    a2->gid = a3->gid;
    a2->size = a3->size;
    a2->blocksize = 512;
    a2->rdev = a3->rdev.specdata1;
    a2->blocks = (a2->size + a2->blocksize - 1) / a2->blocksize;
    a2->fsid = a3->fsid;
    a2->fileid = a3->fileid & 0xfffffff;		/* mask top bits */
    a2->atime.seconds = a3->atime.seconds;
    a2->atime.useconds = (a3->atime.nseconds + 500) / 1000;
    a2->mtime.seconds = a3->mtime.seconds;
    a2->mtime.useconds = (a3->mtime.nseconds + 500) / 1000;
    a2->ctime.seconds = a3->ctime.seconds;
    a2->ctime.useconds = (a3->ctime.nseconds + 500) / 1000;
}


/******************************************************************************
Return pointer to last component of a path. If the path is '/', we'll
return a pointer to a string of zero length, so be prepared to handle this.
You should have already used nfsCleanPath to clean up this path.
******************************************************************************/
filename3 nfsLastComponent (const filename3 path)
{
    char		*s;

    s = strrchr (path, '/');
    if (s == NULL)
	return path;
    return s + 1;
}


/******************************************************************************
Strip repeated / characters and trailing / characters from a path, yielding
a path of components joined by single / characters.
******************************************************************************/
void nfsCleanPath (filename3 path)
{
    char		*s;
    char		*t;
    char		c;

    /* Don't try to clean up an empty string. */
    if (*path == '\0')
	return;

    /* Eliminate redundant / characters. */
    for (s = path + 1, t = path; (c = *s); ++s)
    {
	/* Got / but already on one? */
	if ((c == '/') && (*t == '/'))
	    continue;
	
	/* Copy a character and go forward. */
	*(++t) = c;
    }

    /* Terminate the path. */
    *(++t) = '\0';

    /* Eliminate trailing / characters. Stop before killing the very first
       character (s == path). */
    for (s = t - 1; (s > path) && (*s == '/'); --s)
	*s = '\0';
}


/******************************************************************************
Free an nfsObject.
******************************************************************************/
void nfsFreeObject (nfsObject *f)
{
    if (f->fh) fhfree (f->fh);
    if (f->dh) fhfree (f->dh);
    if (f->attributes) free (f->attributes);
    if (f->name) free (f->name);
    if (f->path) free (f->path);
    free (f);
}


/******************************************************************************
Free a single directory node.
******************************************************************************/
void nfsFreeDirectoryNode (nfsDirectory *dl)
{
    if (dl->object)
	nfsFreeObject (dl->object);
    free (dl);
}


/******************************************************************************
Free a list of directory nodes.
******************************************************************************/
void nfsFreeDirectory (nfsDirectory *dl)
{
    nfsDirectory	*next;

    for ( ; dl; dl = next)
    {
	next = dl->next;
	nfsFreeDirectoryNode (dl);
    }
}


/******************************************************************************
Free a directory path.
******************************************************************************/
void nfsFreePath (nfsPath *path)
{
    nfsPath		*p;

    for ( ; path; path = p)
    {
	p = path->next;
	fhfree (path->fh);
	free (path->name);
	free (path);
    }
}


/******************************************************************************
Find a file from a path. Path may be relative or absolute. If relative,
the file is found relative to the current directory maintained in the NFS
context. If the file is not found, and parent is non-zero, returns a
result containing the filehandle for the parent, with the filehandle for
the requested object set to NULL. In this case, the attributes will also be
NULL.
******************************************************************************/
nfsObject *nfsFind (const filename3 name, const int parent, nfsContext *con)
{
    char		*path;
    char		*s;
    char		*t;
    nfsObject		*f = NULL;
    nfsObject		*d;
    nfs_fh3		*dh;

    /* Need something, dammit. */
    if (*name == '\0')
    {
	nfserrno = (nfsstat3) NFSERR_INVAL;
	return NULL;
    }

    /* Make a copy of the path. */
    path = strdup (name);

    /* Clean the path. */
    nfsCleanPath (path);

    /* Choose start directory (relative or absolute). */
    if (*path == '/')
    {
	/* Start at root and skip initial /. */
	dh = fhdup (con->root->fh);
	s = path + 1;

	/* If that's all there was, we still need to look up something. */
	if (*s == '\0')
	    s = ".";
    }
    else
    {
	/* Start at cwd. */
	dh = fhdup (con->cwd->fh);
	s = path;
    }

    do
    {
	/* Determine next component. */
	if ((t = strchr (s, '/')))
	    *t = '\0';

	/* Try to lookup next component. */
	if (*s)
	{
	    d = f;
	    f = nfsLookup (dh, s, con);
	    if (f == NULL)
	    {
		/* Next component doesn't exist. Was it the last component
		   and was parent set? */
		if ((t == NULL) && parent)
		{
		    /* Pretend we found the object, but leave its filehandle
		       and attributes NULL. */
		    f = (nfsObject *) malloc (sizeof (*f));
		    f->path = path;
		    f->name = strdup (s);
		    f->attributes = NULL;
		    f->dh = dh;
		    f->fh = NULL;

		    return f;
		}

		fhfree (dh);
		free (path);
		return NULL;
	    }

	    /* Free parent object. */
	    if (d)
		nfsFreeObject (d);

	    /* Free parent handle. */
	    fhfree (dh);

	    /* Copy new parent handle. */
	    dh = fhdup (f->fh);
	}

	/* Next component. */
	if (t)
	{
	    *t = '/';
	    s = t + 1;
	}
	else
	    break;
    } while (1);

    /* Done with this. */
    fhfree (dh);

    /* Replace path (last component only) with full path. */
    free (f->path);
    f->path = path;

    return f;
}


/******************************************************************************
Call nfsFind() for a list of paths, and return a sorted directory list
composed of the NFS objects.
******************************************************************************/
nfsDirectory *nfsFindMultiple (size_t argc, char **argv, nfsContext *con)
{
    nfsDirectory	*dl = NULL;
    nfsDirectory	*d;
    nfsObject		*f;
    size_t		k;

    for (k = 0; k < argc; ++k)
    {
	char		*path = argv[k];

	if ((!fhcmp (con->cwd->fh, con->root->fh)) && !strcmp (path, ".."))
	    path = ".";

	/* Lookup target. */
	f = nfsFind (path, 0, con);
	if (f == NULL)
	{
	    nfsperror (path);
	    continue;
	}

	/* Create an entry in our file list. */
	d = (nfsDirectory *) malloc (sizeof (*d));
	d->object = f;
	d->next = dl;
	dl = d;
    }

    /* Sort the list. */
    return nfsSortDirectory (dl);
}


/******************************************************************************
Li'l sort function for qsort. We sort by path, lexicographically.
******************************************************************************/
static int sortfn (const void *a, const void *b)
{
    return strcmp ((*((nfsDirectory **) a))->object->path, (*((nfsDirectory **) b))->object->path);
}


/******************************************************************************
Sort a directory list by path.
******************************************************************************/
nfsDirectory *nfsSortDirectory (nfsDirectory *dl)
{
    nfsDirectory	**names;
    nfsDirectory	**u;
    size_t		count;
    size_t		k;
    nfsDirectory	*p;

    /* We have a list of directory contents. Sort it. */
    if (dl == NULL)
	return NULL;

    /* First find out how long it is. */
    for (p = dl, count = 0; p; p = p->next, ++count)
	;

    /* Create an array of the node pointers. */
    names = (nfsDirectory **) malloc (count * sizeof (*names));
    for (p = dl, u = names; p; p = p->next, ++u)
	*u = p;

    /* Sort it. */
    qsort (names, count, sizeof (names[0]), sortfn);

    /* Recreate list from array. */
    for (k = 0; k < count - 1; ++k)
	names[k]->next = names[k + 1];
    names[k]->next = NULL;
    dl = names[0];

    /* Done with this. */
    free (names);

    return dl;
}


/******************************************************************************
Lookup the filehandle and attributes of a file within a directory. If the
directory handle (dfh) is NULL, lookup the object by treating it as an
explicit filehandle, and treat it as a child of the current filesystem root.
******************************************************************************/
nfsObject *nfsLookup (const nfs_fh3 *dfh, const filename3 target, nfsContext *con)
{
    nfsObject		f;
    fattr3		attr;
    filename3		name = target;

    /* If no parent directory was passed, treat name as an explicit
       filehandle. */
    if (dfh == NULL)
    {
	nfs_fh3		*ph = parseHandle (name);
	nfsObject	*p;

	if (ph == NULL)
	{
	    return NULL;
	}

	/* Pretend we found the object, but leave its filehandle and
	   attributes NULL. Treat the current root as its parent. */
	f.path = strdup (name);
	f.name = strdup (name);
	f.attributes = NULL;
	f.dh = fhdup (con->root->fh);
	f.fh = ph;
	p = memdup (&f, sizeof (f));

	/* Look up attributes for the object. */
	if (nfsGetattr (p, con) < 0)
	{
	    nfsFreeObject (p);
	    return NULL;
	}

	return p;
    }

    if (con->version == NFS_VERSION)
    {
	diropargs	a;
	diropres	*r;

	a.dir = *fh3to2 (dfh);
	a.name = name;
	r = nfsproc_lookup_2 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS_OK)
	    nfserrno = (nfsstat3) r->status;
	else
	{
	    diropokres	*d = &r->diropres_u.diropres;

	    f.path = strdup (name);
	    f.name = strdup (name);
	    attr2to3 (&attr, &d->attributes);
	    f.attributes = memdup (&attr, sizeof (attr));
	    f.dh = fhdup (dfh);
	    f.fh = fh2dup (&d->file);
	    return memdup (&f, sizeof (f));
	}
    }
    else if (con->version == NFS_V3)
    {
	LOOKUP3args	a;
	LOOKUP3res	*r;

	a.what.dir = *dfh;
	a.what.name = name;
	r = nfsproc3_lookup_3 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS3_OK)
	    nfserrno = r->status;
	else
	{
	    LOOKUP3resok	*d = &r->LOOKUP3res_u.resok;

	    if (!d->obj_attributes.attributes_follow)
	    {
		nfserrno = (nfsstat3) NFSERR_IO;
		return NULL;
	    }
	    f.path = strdup (name);
	    f.name = strdup (name);
	    attr = d->obj_attributes.post_op_attr_u.attributes;
	    f.attributes = memdup (&attr, sizeof (attr));
	    f.dh = fhdup (dfh);
	    f.fh = fhdup (&d->object);
	    return memdup (&f, sizeof (f));
	}
    }
    else
    {
	nfserrno = (nfsstat3) NFSERR_OPNOTSUPP;
    }

    return NULL;
}


/******************************************************************************
Read a directory and return a sorted list of its contents. For each entry,
the object is looked up.
******************************************************************************/
nfsDirectory *nfsReaddir (nfs_fh3 *fh, nfsContext *con)
{
    nfsDirectory	*dl = NULL;
    nfsDirectory	*p;
    nfsObject		*f;
    size_t		count = 0;
    int			notRoot = fhcmp (fh, con->root->fh);

    if (con->version == NFS_VERSION)
    {
	readdirargs	a;
	readdirres	*r;
	entry		*e;

	a.dir = *fh3to2 (fh);
	memset (a.cookie, 0, sizeof (a.cookie));
	a.count = NFS_MAXDATA;

	while (1)
	{
	    r = nfsproc_readdir_2 (&a, con->client);
	    if (r == NULL)
	    {
		nfserrno = (nfsstat3) NFSERR_IO;
		break;
	    }
	    else if (r->status != NFS_OK)
	    {
		nfserrno = (nfsstat3) r->status;
		nfsFreeDirectory (dl);
		return NULL;
	    }
	    for (e = r->readdirres_u.reply.entries; e; e = e->nextentry)
	    {
		/* Mask off upper bits of file id. */
		e->fileid &= 0xfffffff;

		/* If not parent of root directory. */
		if (notRoot || strcmp (e->name, ".."))
		{
		    /* Get the file info. */
		    f = nfsLookup (fh, e->name, con);
		    if (f == NULL)
		    {
			fprintf (stderr, "%9u ", e->fileid);
			nfsperror (e->name);
			memcpy (a.cookie, &e->cookie, sizeof (a.cookie));
			continue;
		    }

		    /* Allocate a new directory node. */
		    p = (nfsDirectory *) malloc (sizeof (*p));
		    p->object = f;
		    p->next = dl;
		    dl = p;
		    ++count;
		}

		memcpy (a.cookie, &e->cookie, sizeof (a.cookie));
	    }

	    if (r->readdirres_u.reply.eof)
		break;
	}
    }
    else if (con->version == NFS_V3)
    {
	READDIR3args	a;
	READDIR3res	*r;
	entry3		*e;

	a.dir = *fh;
	memset (&a.cookie, 0, sizeof (a.cookie));
	memset (&a.cookieverf, 0, sizeof (a.cookieverf));
	a.count = NFS_MAXDATA;

	while (1)
	{
	    r = nfsproc3_readdir_3 (&a, con->client);
	    if (r == NULL)
	    {
		nfserrno = (nfsstat3) NFSERR_IO;
		break;
	    }
	    else if (r->status != NFS3_OK)
	    {
		nfserrno = r->status;
		nfsFreeDirectory (dl);
		return NULL;
	    }
	    for (e = r->READDIR3res_u.resok.reply.entries; e; e = e->nextentry)
	    {
		/* Mask off upper bits of file id. */
		e->fileid &= 0xfffffff;

		/* If not parent of root directory. */
		if (notRoot || strcmp (e->name, ".."))
		{
		    /* Get the file info. */
		    f = nfsLookup (fh, e->name, con);
		    if (f == NULL)
		    {
			fprintf (stderr, "%9" PRINTF_INT64_FORMAT "u ", e->fileid);
			nfsperror (e->name);
			memcpy (&a.cookie, &e->cookie, sizeof (a.cookie));
			continue;
		    }
		}
		else
		    continue;

		/* Allocate a new directory node. */
		p = (nfsDirectory *) malloc (sizeof (*p));
		p->object = f;
		p->next = dl;
		dl = p;
		memcpy (&a.cookie, &e->cookie, sizeof (a.cookie));
		++count;
	    }
	    memcpy (&a.cookieverf, &r->READDIR3res_u.resok.cookieverf, sizeof (a.cookieverf));

	    if (r->READDIR3res_u.resok.reply.eof)
		break;
	}
    }

    return nfsSortDirectory (dl);
}


/******************************************************************************
Get attributes of an NFS object.
******************************************************************************/
int nfsGetattr (nfsObject *f, nfsContext *con)
{
    fattr3		attr;

    if (con->version == NFS_VERSION)
    {
	attrstat	*r;
	nfs_fh		fh;

	fh = *fh3to2 (f->fh);

	r = nfsproc_getattr_2 (&fh, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS_OK)
	    nfserrno = (nfsstat3) r->status;
	else
	{
	    attr2to3 (&attr, &r->attrstat_u.attributes);
	    if (f->attributes)
		*f->attributes = attr;
	    else
		f->attributes = memdup (&attr, sizeof (attr));
	    return 0;
	}
    }
    else if (con->version == NFS_V3)
    {
	GETATTR3args	a;
	GETATTR3res	*r;

	a.object = *f->fh;
	r = nfsproc3_getattr_3 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS3_OK)
	    nfserrno = r->status;
	else
	{
	    attr = r->GETATTR3res_u.resok.obj_attributes;
	    if (f->attributes)
		*f->attributes = attr;
	    else
		f->attributes = memdup (&attr, sizeof (attr));
	    return 0;
	}
    }
    else
    {
	nfserrno = (nfsstat3) NFSERR_OPNOTSUPP;
    }

    return -1;
}


/******************************************************************************
Set attributes on an NFS object. The what parameter is a bitmask of
attributes to be set. Attributes not in the mask are left intact. The new
attributes are copied out of the nfsObject's attributes structure. So, to
change attributes, look up the object, change its attributes, and call
nfsSetattr() with the attributes you want to change.
******************************************************************************/
int nfsSetattr (const nfsObject *f, const uint32 what, nfsContext *con)
{
    if (con->version == NFS_VERSION)
    {
	sattrargs	a;
	attrstat	*r;
	fattr		attr;
	static nfstime	never = { -1, -1 }; /* Not clear what we need here. */

	a.file = *fh3to2 (f->fh);
	attr3to2 (&attr, f->attributes);
	a.attributes.mode = (what & NFS_ATTR_MODE) ? attr.mode : -1;
	a.attributes.uid = (what & NFS_ATTR_UID) ? attr.uid : -1;
	a.attributes.gid = (what & NFS_ATTR_GID) ? attr.gid : -1;
	a.attributes.size = (what & NFS_ATTR_SIZE) ? attr.size : -1;
	a.attributes.atime = (what & NFS_ATTR_ATIME) ? attr.atime : never;
	a.attributes.mtime = (what & NFS_ATTR_MTIME) ? attr.mtime : never;

	r = nfsproc_setattr_2 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS_OK)
	    nfserrno = (nfsstat3) r->status;
	else
	    return 0;
    }
    else if (con->version == NFS_V3)
    {
	SETATTR3args	a;
	SETATTR3res	*r;
	fattr3		attr;

	a.object = *f->fh;
	attr = *f->attributes;
	a.new_attributes.mode.set_it = (what & NFS_ATTR_MODE) ? 1 : 0;
	a.new_attributes.mode.set_mode3_u.mode = attr.mode;
	a.new_attributes.uid.set_it = (what & NFS_ATTR_UID) ? 1 : 0;
	a.new_attributes.uid.set_uid3_u.uid = attr.uid;
	a.new_attributes.gid.set_it = (what & NFS_ATTR_GID) ? 1 : 0;
	a.new_attributes.gid.set_gid3_u.gid = attr.gid;
	a.new_attributes.size.set_it = (what & NFS_ATTR_SIZE) ? 1 : 0;
	a.new_attributes.size.set_size3_u.size = attr.size;
	a.new_attributes.mtime.set_it = (what & NFS_ATTR_MTIME) ? SET_TO_CLIENT_TIME : DONT_CHANGE;
	a.new_attributes.mtime.set_mtime_u.mtime = attr.mtime;
	a.new_attributes.atime.set_it = (what & NFS_ATTR_ATIME) ? SET_TO_CLIENT_TIME : DONT_CHANGE;
	a.new_attributes.atime.set_atime_u.atime = attr.atime;
	a.guard.check = 0;

	r = nfsproc3_setattr_3 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS3_OK)
	    nfserrno = r->status;
	else
	    return 0;
    }
    else
    {
	nfserrno = (nfsstat3) NFSERR_OPNOTSUPP;
    }

    return -1;
}


/******************************************************************************
Remove an NFS object.
******************************************************************************/
int nfsRemove (const nfsObject *f, nfsContext *con)
{
    if (con->version == NFS_VERSION)
    {
	diropargs	a;
	nfsstat		*r;

	a.dir = *fh3to2 (f->dh);
	a.name = f->name;
	r = nfsproc_remove_2 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (*r != NFS_OK)
	    nfserrno = (nfsstat3) *r;
	else
	    return 0;
    }
    else if (con->version == NFS_V3)
    {
	REMOVE3args	a;
	REMOVE3res	*r;

	a.object.dir = *f->dh;
	a.object.name = f->name;
	r = nfsproc3_remove_3 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS3_OK)
	    nfserrno = r->status;
	else
	    return 0;
    }
    else
    {
	nfserrno = (nfsstat3) NFSERR_OPNOTSUPP;
    }

    return -1;
}


/******************************************************************************
Create an NFS file object. The object's permissions are 0666 modified by
the umask in the nfsContext structure.
******************************************************************************/
int nfsCreate (const nfs_fh3 *fh, const filename3 name, nfsContext *con)
{
    if (con->version == NFS_VERSION)
    {
	createargs	a;
	diropres	*r;

	a.where.dir = *fh3to2 (fh);
	a.where.name = name;
	a.attributes.mode = (0666 & (~(con->umask)));
	a.attributes.uid = (u_int) con->authUid;
	a.attributes.gid = (u_int) con->authGid;
	a.attributes.size = 0;
	a.attributes.atime.seconds = nfsGetAtime (con);
	a.attributes.mtime.seconds = nfsGetMtime (con);
	a.attributes.mtime.useconds = a.attributes.atime.useconds = 0;

	r = nfsproc_create_2 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS_OK)
	    nfserrno = (nfsstat3) r->status;
	else
	    return 0;
    }
    else if (con->version == NFS_V3)
    {
	CREATE3args	a;
	CREATE3res	*r;

	a.where.dir = *fh;
	a.where.name = name;
	a.how.mode = UNCHECKED;
	a.how.createhow3_u.obj_attributes.mode.set_it = 1;
	a.how.createhow3_u.obj_attributes.mode.set_mode3_u.mode = (0666 & (~(con->umask)));
	a.how.createhow3_u.obj_attributes.uid.set_it = 1;
	a.how.createhow3_u.obj_attributes.uid.set_uid3_u.uid = (u_int) con->authUid;
	a.how.createhow3_u.obj_attributes.gid.set_it = 1;
	a.how.createhow3_u.obj_attributes.gid.set_gid3_u.gid = (u_int) con->authGid;
	a.how.createhow3_u.obj_attributes.size.set_it = 0;
	a.how.createhow3_u.obj_attributes.atime.set_it = a.how.createhow3_u.obj_attributes.mtime.set_it = SET_TO_CLIENT_TIME;
	a.how.createhow3_u.obj_attributes.atime.set_atime_u.atime.seconds = nfsGetAtime (con);
	a.how.createhow3_u.obj_attributes.mtime.set_mtime_u.mtime.seconds = nfsGetMtime (con);

	r = nfsproc3_create_3 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS3_OK)
	    nfserrno = r->status;
	else
	    return 0;
    }
    else
    {
	nfserrno = (nfsstat3) NFSERR_OPNOTSUPP;
    }

    return -1;
}


/******************************************************************************
Remove a directory.
******************************************************************************/
int nfsRmdir (const nfsObject *f, nfsContext *con)
{
    if (con->version == NFS_VERSION)
    {
	diropargs	a;
	nfsstat		*r;

	a.dir = *fh3to2 (f->dh);
	a.name = f->name;
	r = nfsproc_rmdir_2 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (*r != NFS_OK)
	{
	    nfserrno = (nfsstat3) *r;
	    if (nfserrno == NFS3ERR_EXIST)
		nfserrno = NFS3ERR_NOTEMPTY;
	}
	else
	    return 0;
    }
    else if (con->version == NFS_V3)
    {
	RMDIR3args	a;
	RMDIR3res	*r;

	a.object.dir = *f->dh;
	a.object.name = f->name;
	r = nfsproc3_rmdir_3 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS3_OK)
	{
	    nfserrno = r->status;
	    if (nfserrno == NFS3ERR_EXIST)
		nfserrno = NFS3ERR_NOTEMPTY;
	}
	else
	    return 0;
    }
    else
    {
	nfserrno = (nfsstat3) NFSERR_OPNOTSUPP;
    }

    return -1;
}


/******************************************************************************
Create an NFS directory. The permissions of the directory will be 0777
modified by the umask in the nfsContext structure.
******************************************************************************/
int nfsMkdir (const nfs_fh3 *fh, const filename3 name, nfsContext *con)
{
    if (con->version == NFS_VERSION)
    {
	createargs	a;
	diropres	*r;

	a.where.dir = *fh3to2 (fh);
	a.where.name = name;
	a.attributes.mode = (0777 & (~(con->umask)));
	a.attributes.uid = (u_int) con->authUid;
	a.attributes.gid = (u_int) con->authGid;
	a.attributes.size = 0;
	a.attributes.mtime.seconds = a.attributes.atime.seconds = time (0);
	a.attributes.mtime.useconds = a.attributes.atime.useconds = 0;

	r = nfsproc_mkdir_2 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS_OK)
	    nfserrno = (nfsstat3) r->status;
	else
	    return 0;
    }
    else if (con->version == NFS_V3)
    {
	MKDIR3args	a;
	MKDIR3res	*r;

	a.where.dir = *fh;
	a.where.name = name;
	a.attributes.mode.set_it = 1;
	a.attributes.mode.set_mode3_u.mode = (0777 & (~(con->umask)));
	a.attributes.uid.set_it = 1;
	a.attributes.uid.set_uid3_u.uid = (u_int) con->authUid;
	a.attributes.gid.set_it = 1;
	a.attributes.gid.set_gid3_u.gid = (u_int) con->authGid;
	a.attributes.size.set_it = 0;
	a.attributes.mtime.set_it = DONT_CHANGE;
	a.attributes.atime.set_it = DONT_CHANGE;

	r = nfsproc3_mkdir_3 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS3_OK)
	    nfserrno = r->status;
	else
	    return 0;
    }
    else
    {
	nfserrno = (nfsstat3) NFSERR_OPNOTSUPP;
    }

    return -1;
}


/******************************************************************************
Copy an NFS object to a stream.
******************************************************************************/
int64 nfsRead (nfsObject *f, FILE *stream, nfsContext *con)
{
    int64		total = 0;
    count3		len;

    if (con->version == NFS_VERSION)
    {
	readargs	a;
	readres		*r;

	a.file = *fh3to2 (f->fh);
	a.offset = 0;
	a.count = NFS_MAXDATA;
	a.totalcount = 0;	/* IGNORED */
	while (1)
	{
	    r = nfsproc_read_2 (&a, con->client);
	    if (r == NULL)
	    {
		nfserrno = (nfsstat3) NFSERR_IO;
		return -1;
	    }
	    else if (r->status != NFS_OK)
	    {
		nfserrno = (nfsstat3) r->status;
		return -1;
	    }

	    /* Count bytes read. */
	    len = r->readres_u.reply.data.data_len;
	    if (len == 0)
		return total;

	    /* Write data to stream. */
	    fwrite (r->readres_u.reply.data.data_val, 1, len, stream);
	    a.offset += len;
	    total += len;
	}
    }
    else if (con->version == NFS_V3)
    {
	READ3args	a;
	READ3res	*r;

	a.file = *f->fh;
	a.offset = 0;
	a.count = NFS_MAXDATA;
	while (1)
	{
	    r = nfsproc3_read_3 (&a, con->client);
	    if (r == NULL)
	    {
		nfserrno = (nfsstat3) NFSERR_IO;
		return -1;
	    }
	    else if (r->status != NFS3_OK)
	    {
		nfserrno = (nfsstat3) r->status;
		return -1;
	    }

	    /* Count bytes read. */
	    len = r->READ3res_u.resok.data.data_len;
	    if (len == 0)
		return total;

	    /* Write data to stream. */
	    fwrite (r->READ3res_u.resok.data.data_val, 1, len, stream);
	    a.offset += len;
	    total += len;
	}
    }
    else
    {
	nfserrno = (nfsstat3) NFSERR_OPNOTSUPP;
    }

    return -1;
}


/******************************************************************************
Copy a stream to an NFS object.
******************************************************************************/
int64 nfsWrite (nfsObject *f, FILE *stream, nfsContext *con)
{
    int64		total = 0;
    count3		len;

    if (con->version == NFS_VERSION)
    {
	writeargs	a;
	attrstat	*r;

	a.file = *fh3to2 (f->fh);
	a.beginoffset = 0;	/* IGNORED */
	a.offset = 0;
	a.totalcount = 0;	/* IGNORED */
	while (1)
	{
	    char	buf[1024];	/* This is just a reasonable value. */

	    len = fread (buf, 1, sizeof (buf), stream);

	    if (len == 0)
	    {
		if (ferror (stream))
		    return -1;

		if (feof (stream))
		    return total;
	    }

	    a.data.data_len = len;
	    a.data.data_val = buf;
	    r = nfsproc_write_2 (&a, con->client);
	    if (r == NULL)
	    {
		nfserrno = (nfsstat3) NFSERR_IO;
		return -1;
	    }
	    else if (r->status != NFS_OK)
	    {
		nfserrno = (nfsstat3) r->status;
		return -1;
	    }

	    /* Count bytes written. */
	    a.offset += len;
	    total += len;
	}
    }
    else if (con->version == NFS_V3)
    {
	WRITE3args	a;
	WRITE3res	*r;

	a.file = *f->fh;
	a.offset = 0;
	a.stable = FILE_SYNC;
	while (1)
	{
	    char	buf[NFS_MAXDATA];
	    char	*s;

	    len = fread (buf, 1, sizeof (buf), stream);

	    if (len == 0)
	    {
		if (ferror (stream))
		    return -1;

		if (feof (stream))
		    return total;
	    }

	    /* May need to write in multiple chunks. */
	    for (s = buf; len > 0; )
	    {
		size_t	wlen;

		a.count = len;
		a.data.data_len = len;
		a.data.data_val = s;
		r = nfsproc3_write_3 (&a, con->client);
		if (r == NULL)
		{
		    nfserrno = (nfsstat3) NFSERR_IO;
		    return -1;
		}
		else if (r->status != NFS3_OK)
		{
		    nfserrno = r->status;
		    return -1;
		}

		/* Count bytes written. */
		wlen = r->WRITE3res_u.resok.count;
		a.offset += wlen;
		total += wlen;
		len -= wlen;
		s += wlen;
	    }
	}
    }
    else
    {
	nfserrno = (nfsstat3) NFSERR_OPNOTSUPP;
    }

    return -1;
}


/******************************************************************************
Read a link.
******************************************************************************/
nfspath3 nfsReadlink (nfs_fh3 *fh, nfsContext *con)
{
    if (con->version == NFS_VERSION)
    {
	readlinkres		*r;

	r = nfsproc_readlink_2 (fh3to2 (fh), con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS_OK)
	    nfserrno = (nfsstat3) r->status;
	else
	    return strdup (r->readlinkres_u.data);
    }
    else if (con->version == NFS_V3)
    {
	READLINK3args		a;
	READLINK3res		*r;

	a.symlink = *fh;
	r = nfsproc3_readlink_3 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS3_OK)
	    nfserrno = r->status;
	else
	    return strdup (r->READLINK3res_u.resok.data);
    }
    else
    {
	nfserrno = (nfsstat3) NFSERR_OPNOTSUPP;
    }

    return NULL;
}


/******************************************************************************
Create a hard link.
******************************************************************************/
int nfsLink (const nfs_fh3 *fh, const filename3 from, const nfs_fh3 *to, nfsContext *con)
{
    if (con->version == NFS_VERSION)
    {
	linkargs	a;
	nfsstat		*r;

	a.from = *fh3to2 (to);
	a.to.dir = *fh3to2 (fh);
	a.to.name = from;

	r = nfsproc_link_2 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (*r != NFS_OK)
	    nfserrno = (nfsstat3) *r;
	else
	    return 0;
    }
    else if (con->version == NFS_V3)
    {
	LINK3args	a;
	LINK3res	*r;

	a.file = *to;
	a.link.dir = *fh;
	a.link.name = from;

	r = nfsproc3_link_3 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS3_OK)
	    nfserrno = r->status;
	else
	    return 0;
    }
    else
    {
	nfserrno = (nfsstat3) NFSERR_OPNOTSUPP;
    }

    return -1;
}


/******************************************************************************
Create a symbolic link.
******************************************************************************/
int nfsSymlink (const nfs_fh3 *fh, const filename3 from, const filename3 to, nfsContext *con)
{
    if (con->version == NFS_VERSION)
    {
	symlinkargs	a;
	nfsstat		*r;

	a.from.dir = *fh3to2 (fh);
	a.from.name = from;
	a.to = to;
	a.attributes.mode = 0777;
	a.attributes.uid = con->authUid;
	a.attributes.gid = con->authGid;
	a.attributes.size = 0;
	a.attributes.atime.seconds = nfsGetAtime (con);
	a.attributes.mtime.seconds = nfsGetMtime (con);
	a.attributes.atime.useconds = a.attributes.mtime.useconds = 0;

	r = nfsproc_symlink_2 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (*r != NFS_OK)
	    nfserrno = (nfsstat3) *r;
	else
	    return 0;
    }
    else if (con->version == NFS_V3)
    {
	SYMLINK3args	a;
	SYMLINK3res	*r;

	a.where.dir = *fh;
	a.where.name = from;
	a.symlink.symlink_attributes.mode.set_it = 1;
	a.symlink.symlink_attributes.mode.set_mode3_u.mode = 0777;
	a.symlink.symlink_attributes.uid.set_it = 1;
	a.symlink.symlink_attributes.uid.set_uid3_u.uid = (u_int) con->authUid;
	a.symlink.symlink_attributes.gid.set_it = 1;
	a.symlink.symlink_attributes.gid.set_gid3_u.gid = (u_int) con->authGid;
	a.symlink.symlink_attributes.size.set_it = 0;
	a.symlink.symlink_attributes.atime.set_it = a.symlink.symlink_attributes.mtime.set_it = SET_TO_CLIENT_TIME;
	a.symlink.symlink_attributes.atime.set_atime_u.atime.seconds = nfsGetAtime (con);
	a.symlink.symlink_attributes.mtime.set_mtime_u.mtime.seconds = nfsGetMtime (con);
	a.symlink.symlink_attributes.atime.set_atime_u.atime.nseconds = a.symlink.symlink_attributes.mtime.set_mtime_u.mtime.nseconds = 0;
	a.symlink.symlink_data = to;

	r = nfsproc3_symlink_3 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS3_OK)
	    nfserrno = r->status;
	else
	    return 0;
    }
    else
    {
	nfserrno = (nfsstat3) NFSERR_OPNOTSUPP;
    }

    return -1;
}


/******************************************************************************
Rename an NFS object.
******************************************************************************/
int nfsRename (const nfs_fh3 *sdh, const filename3 sname, const nfs_fh3 *ddh, const filename3 dname, nfsContext *con)
{
    if (con->version == NFS_VERSION)
    {
	renameargs	a;
	nfsstat		*r;

	a.from.dir = *fh3to2 (sdh);
	a.from.name = sname;
	a.to.dir = *fh3to2 (ddh);
	a.to.name = dname;

	r = nfsproc_rename_2 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (*r != NFS_OK)
	    nfserrno = (nfsstat3) *r;
	else
	    return 0;
    }
    else if (con->version == NFS_V3)
    {
	RENAME3args	a;
	RENAME3res	*r;

	a.from.dir = *sdh;
	a.from.name = sname;
	a.to.dir = *ddh;
	a.to.name = dname;

	r = nfsproc3_rename_3 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS3_OK)
	    nfserrno = r->status;
	else
	    return 0;
    }
    else
    {
	nfserrno = (nfsstat3) NFSERR_OPNOTSUPP;
    }

    return -1;
}


/******************************************************************************
Get NFS3 static filesystem info.
******************************************************************************/
FSINFO3resok *nfsFsinfo (const nfs_fh3 *fh, nfsContext *con)
{
    if (con->version == NFS_V3)
    {
	FSINFO3args	a;
	FSINFO3res	*r;

	a.fsroot = *fh;
	r = nfsproc3_fsinfo_3 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS3_OK)
	    nfserrno = r->status;
	else
	    return memdup (&r->FSINFO3res_u.resok, sizeof (r->FSINFO3res_u.resok));
    }
    else
    {
	nfserrno = (nfsstat3) NFSERR_OPNOTSUPP;
    }

    return NULL;
}


/******************************************************************************
Get dynamic filesystem info.
******************************************************************************/
FSSTAT3resok *nfsFsstat (const nfs_fh3 *fh, nfsContext *con)
{
    if (con->version == NFS_VERSION)
    {
	statfsres	*r;
	FSSTAT3resok	f;

	r = nfsproc_statfs_2 (fh3to2 (fh), con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS_OK)
	    nfserrno = (nfsstat3) r->status;
	else
	{
	    size3	bs = r->statfsres_u.reply.bsize;

	    f.obj_attributes.attributes_follow = 0;
	    f.tbytes = bs * (size3) r->statfsres_u.reply.blocks;
	    f.fbytes = bs * (size3) r->statfsres_u.reply.bfree;
	    f.abytes = bs * (size3) r->statfsres_u.reply.bavail;
	    f.tfiles = 0;
	    f.ffiles = 0;
	    f.afiles = 0;
	    f.invarsec = 0;
	    return memdup (&f, sizeof (f));
	}
    }
    else if (con->version == NFS_V3)
    {
	FSSTAT3args	a;
	FSSTAT3res	*r;

	a.fsroot = *fh;
	r = nfsproc3_fsstat_3 (&a, con->client);
	if (r == NULL)
	    nfserrno = (nfsstat3) NFSERR_IO;
	else if (r->status != NFS3_OK)
	    nfserrno = r->status;
	else
	    return memdup (&r->FSSTAT3res_u.resok, sizeof (r->FSSTAT3res_u.resok));
    }
    else
    {
	nfserrno = (nfsstat3) NFSERR_OPNOTSUPP;
    }

    return NULL;
}


/******************************************************************************
Create new RPC authentication structure.
******************************************************************************/
static void authChange (nfsContext *con)
{
    if (con->client->cl_auth)
	auth_destroy (con->client->cl_auth);
    con->client->cl_auth = authunix_create (con->authHost, con->authUid, con->authGid, 1, &con->authGid);
}


/******************************************************************************
Create a new NFS context with its root and cwd set to the given filehandle.
The credentials and umask will be set to those of the current process.
******************************************************************************/
nfsContext *nfsNewContext (const nfs_fh3 *root, CLIENT *client, const u_long version)
{
    nfsContext		*new;

    new = (nfsContext *) malloc (sizeof (*new));

    /* Initialize RPC info. */
    new->client = client;
    new->version = version;

    /* Initialize root and cwd. */
    new->root = (nfsPath *) malloc (sizeof (*new->root));
    new->root->fh = fhdup (root);
    new->root->name = strdup ("");
    new->root->parent = new->root;
    new->root->next = NULL;
    new->cwd = new->root;

    /* Initialize credentials. */
    new->umask = umask (0);	/* This kills the process umask. */
    umask (new->umask);		/* Restore process umask. */
    new->authHost = strdup ("localhost");
    new->authUid = getuid ();
    new->authGid = getgid ();

    /* Create initial RPC authentication structure. */
    authChange (new);

    /* Ignore atime and mtime in context. */
    new->useAtime = new->useMtime = 0;
    new->atime = new->mtime = time (0);

    return new;
}


/******************************************************************************
Get path to cwd.
******************************************************************************/
filename3 nfsGetPath (const nfsContext *con)
{
    nfsPath		*p;
    char		*path;
    char		*s;
    char		*t;
    size_t		len;

    for (p = con->root, len = 1; p; p = p->next)
    {
	len += strlen (p->name);
	if (p->next)
	    ++len;
    }

    s = path = malloc (len + 1);
    *(s++) = '/';
    for (p = con->root->next; p; p = p->next)
    {
	for (t = p->name; (*s = *t); ++t, ++s)
	    ;
	if (p->next)
	    *(s++) = '/';
    }
    *s = '\0';

    return path;
}


/******************************************************************************
Change uid and create new authorization.
******************************************************************************/
void nfsSetUser (const uid_t uid, nfsContext *con)
{
    con->authUid = uid;
    authChange (con);
}


/******************************************************************************
Change gid and create new authorization.
******************************************************************************/
void nfsSetGroup (const gid_t gid, nfsContext *con)
{
    con->authGid = gid;
    authChange (con);
}


/******************************************************************************
Set or clear context atime. If time is NULL, clears useAtime flag.
******************************************************************************/
void nfsSetAtime (const time_t *time, nfsContext *con)
{
    if (time)
    {
	con->atime = *time;
	con->useAtime = 1;
    }
    else
	con->useAtime = 0;
}


/******************************************************************************
Set or clear context mtime. If time is NULL, clears useMtime flag.
******************************************************************************/
void nfsSetMtime (const time_t *time, nfsContext *con)
{
    if (time)
    {
	con->mtime = *time;
	con->useMtime = 1;
    }
    else
	con->useMtime = 0;
}


/******************************************************************************
Set cwd to requested path. 
******************************************************************************/
int nfsChdir (const filename3 name, nfsContext *con)
{
    char		*path = strdup (name);
    char		*s;
    char		*t;
    nfsPath		*cwd;
    nfsPath		*attach = NULL;
    nfsPath		*branch = NULL;
    nfsPath		*p;

    /*nfsCleanPath (path);*/

    /* Relative or absolute? */
    if (*path == '/')
	cwd = con->root;
    else
	cwd = con->cwd;

    for (s = path, t = strchr (s, '/'); ; )
    {
	if (t)
	    *t = '\0';
	if ((*s == '\0') || !strcmp (s, "."))
	    ;
	else if (!strcmp (s, ".."))
	{
	    /* Go up one. */
	    cwd = cwd->parent;

	    /* If we are nowhere we must have ascended the branch. */
	    if (cwd == NULL)
	    {
		/* Return to the main path and free the branch. */
		cwd = attach;
		nfsFreePath (branch);
		branch = NULL;
		attach = NULL;
	    }
	}
	else
	{
	    nfsObject	*f;

	    f = nfsLookup (cwd->fh, s, con);
	    if (f == NULL)
	    {
		/* Can't search component, so give up. */
		nfsFreePath (branch);
		free (path);
		return -1;
	    }

	    if (f->attributes->type != NF3DIR)
	    {
		nfserrno = NFS3ERR_NOTDIR;
		nfsFreeObject (f);
		nfsFreePath (branch);
		free (path);
		return -1;
	    }

	    /* Create new path node. */
	    p = (nfsPath *) malloc (sizeof (*p));
	    p->fh = fhdup (f->fh);
	    p->name = strdup (s);
	    p->parent = cwd;
	    p->next = NULL;

	    /* If no branch yet, create one, and remember where to attach
	       it later. */
	    if (branch == NULL)
	    {
		attach = cwd;
		branch = p;
		p->parent = NULL;
	    }
	    else
	    {
		/* Not on main branch, so attach new node to cwd. */
		cwd->next = p;
	    }

	    /* Set new cwd. */
	    cwd = p;

	    nfsFreeObject (f);
	}

	if (t)
	{
	    s = t + 1;
	    t = strchr (s, '/');
	}
	else
	    break;
    }

    /* Free everything below the new cwd. */
    nfsFreePath (cwd->next);
    cwd->next = NULL;

    /* If no attachment point, cwd must be on the main path, so we'll use
       that. */
    if (attach == NULL)
	attach = cwd;

    /* We made it. Now attach the branch to the main path, and free the
       old path section. */
    nfsFreePath (attach->next);
    attach->next = branch;
    if (branch)
	branch->parent = attach;

    /* And set the new cwd. */
    con->cwd = cwd;

    free (path);
    return 0;
}


/******************************************************************************
Set cwd to requested path. 
******************************************************************************/
int nfsChroot (const filename3 name, nfsContext *con)
{
    nfsObject		*f;
    nfs_fh3		*ph = NULL;

    /*nfsCleanPath (path);*/

    /* We don't need to do anything fancy here, because we're going to throw
       away the old root and cwd paths. */

    /* First check for ##filehandle as argument. */
    if ((*(name + 0) == '#') && (*(name + 1) == '#') && ((ph = parseHandle (name + 2))))
    {
	/* Lookup the explicit filehandle. */
	f = nfsLookup (NULL, name + 2, con);
    }
    else
    {
	/* Find the directory by path. */
	f = nfsFind (name, 0, con);
    }
    if (f == NULL)
    {
	if (ph)
	    fhfree (ph);
	return -1;
    }
    if (f->attributes->type != NF3DIR)
    {
	nfserrno = NFS3ERR_NOTDIR;
	nfsFreeObject (f);
	return -1;
    }

    /* Free the old root path. */
    nfsFreePath (con->root);

    /* Create new path node. */
    con->root = (nfsPath *) malloc (sizeof (*con->root));
    con->root->fh = fhdup (f->fh);
    con->root->name = strdup ("");
    con->root->parent = con->root;
    con->root->next = NULL;
    con->cwd = con->root;

    /* Done with this. */
    nfsFreeObject (f);

    return 0;
}


