
/******************************************************************************

Simple NFS client

Author: Jefferson Ogata (JO317) <ogata@pobox.com>
Date: 1999/06/07
Version: 0.1

Please note that this program comes with NO WARRANTY WHATSOEVER. Your use
of this program constitutes your complete acceptance of all liability for
any damage or loss caused by said use.

You are free to use or redistribute this program as long as this header
remains intact.

If you have suggestions -- or better yet -- bits of new code, send them
to me and I will deal with them when I have time. The latest version of
this program may always be found at the URL

    http://pobox.com/~ogata/webtools/nfs

nfsprocs.h: definitions for version-independent wrappers for NFS RPCs.

Change history:

******************************************************************************/

#ifndef	_NFSPROCS_H
#define	_NFSPROCS_H


#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "nfs_prot.h"


/* Structure for maintaining cwd. */
struct nfsPath
{
    nfs_fh3			*fh;
    filename3			name;
    struct nfsPath		*parent;
    struct nfsPath		*next;
};
typedef struct nfsPath		nfsPath;


/* All of the RPC-related information, along with our "process" context:
   credentials, cwd, root directory, and umask. */
struct nfsContext
{
    CLIENT			*client;
    u_long			version;
    nfsPath			*root;
    nfsPath			*cwd;
    char			*authHost;
    uid_t			authUid;
    gid_t			authGid;
    mode_t			umask;
    int				useAtime;
    time_t			atime;
    int				useMtime;
    time_t			mtime;
};
typedef struct nfsContext	nfsContext;


/* An NFS object. Under certain circumstances, attributes and fh may be NULL.
   See nfsFind() for more info. */
struct nfsObject
{
    filename3			path;		/* path from cwd to object */
    filename3			name;		/* last component of path */
    fattr3			*attributes;	/* object attributes */
    nfs_fh3			*dh;		/* parent filehandle */
    nfs_fh3			*fh;		/* object filehandle */
};
typedef struct nfsObject	nfsObject;


/* Linked list of objects. */
struct nfsDirectory
{
    nfsObject			*object;
    struct nfsDirectory		*next;
};
typedef struct nfsDirectory	nfsDirectory;


/* Bits to indicate things for nfsSetattr to set. */
typedef enum nfsAttributes
{
    NFS_ATTR_MODE = 1,
    NFS_ATTR_UID = 2,
    NFS_ATTR_GID = 4,
    NFS_ATTR_SIZE = 8,
    NFS_ATTR_ATIME = 16,
    NFS_ATTR_MTIME = 32,
}				nfsAttributes;


/* Status of the last unsuccessful NFS call. Usually comes from the server,
   but there are a few cases where we make one up. */
extern nfsstat3 nfserrno;

/* strerror for NFS. */
char *nfserror (const nfsstat3 status);

/* perror for NFS. */
void nfsperror (const char *msg);


/* Free a version 3 filehandle. Note we never do anything to malloc
   version 2 handles. We just copy them into version 3 structures. The
   fh3to2() function exists solely for filling in the data structures we
   pass to version 2 RPCs. */
void fhfree (nfs_fh3 *fh);

/* Duplicate a version 3 filehandle. When you are done with it, pass the
   duplicate to fhfree(). */
nfs_fh3 *fhdup (const nfs_fh3 *fh);

/* Duplicate a version 2 filehandle into a version 3 handle. When you are
   done with it, pass the duplicate to fhfree(). */
nfs_fh3 *fh2dup (const nfs_fh *fh);

/* Get pointer to version 2 part of version 3 filehandle. Will ABORT THE
   PROGRAM if the version 3 handle is not really a version 2 handle! */
nfs_fh *fh3to2 (const nfs_fh3 *fh);

/* Compare two version 3 filehandles a la strcmp. */
int fhcmp (const nfs_fh3 *a, const nfs_fh3 *b);


/* Get a pointer to the last component of path. Duh. Does not modify path. */
filename3 nfsLastComponent (const filename3 path);

/* Clean extra / characters out of a path. Doesn't do fancy interpretation
   of "." or "..". Modifies path in place, possibly shortening it. */
void nfsCleanPath (filename3 path);


/* Free an nfsObject. */
void nfsFreeObject (nfsObject *f);

/* Free a single directory list node. */
void nfsFreeDirectoryNode (nfsDirectory *dl);

/* Free a directory list. */
void nfsFreeDirectory (nfsDirectory *dl);

/* Free a directory path. */
void nfsFreePath (nfsPath *p);


/* Find an object. Interprets the path in the current context, looking up
   each component in order. If the object is found it is returned. If the
   object is not found, but the entire path up until the last component
   could be looked up, AND if parent is non-zero, returns an nfsObject with
   a NULL pointer in the attributes and fh fields. The dh, path, and name
   fields will all be valid, and may be used in creating the desired object.
   Returns NULL on failure. */
nfsObject *nfsFind (const filename3 path, const int parent, nfsContext *con);

/* Call nfsFind() for a list of files and return in form of a sorted
   directory listing. */
nfsDirectory *nfsFindMultiple (size_t argc, char **argv, nfsContext *con);

/* Sort a directory. The original structures are relinked into a new order. */
nfsDirectory *nfsSortDirectory (nfsDirectory *dl);


/* Look up an object. The object name is interpreted not as a path, but
   merely an entry in the specified directory. As a special case, the string
   '##' followed by a filehandle is interpreted as a filehandle to look up;
   the directory is then irrelevant. When you are done with it, pass the
   returned nfsObject to nfsFreeObject(). Returns NULL on failure. */
nfsObject *nfsLookup (const nfs_fh3 *dir, const filename3 target, nfsContext *con);

/* Read a directory and return a sorted list of its contents. Returns NULL on
   failure. */
nfsDirectory *nfsReaddir (nfs_fh3 *dir, nfsContext *con);

/* Get file attributes. Returns 0 on success, -1 on failure. */
int nfsGetattr (nfsObject *f, nfsContext *con);

/* Set file attributes. The attributes specified by the what bitmask are
   modified. Returns 0 on success, -1 on failure. */
int nfsSetattr (const nfsObject *f, const uint32 what, nfsContext *con);

/* Remove an NFS object. Returns 0 on success, -1 on failure. */
int nfsRemove (const nfsObject *f, nfsContext *con);

/* Create an NFS file object. The object's permissions are 0666 modified
   by the umask in the nfsContext structure. Returns 0 on success, -1 on
   failure. */
int nfsCreate (const nfs_fh3 *dir, const filename3 name, nfsContext *con);

/* Remove a directory. Returns 0 on success, -1 on failure. */
int nfsRmdir (const nfsObject *f, nfsContext *con);

/* Create a directory. The directory's permissions are 0777 modified by the
   umask in the nfsContext structure. Returns 0 on success, -1 on failure. */
int nfsMkdir (const nfs_fh3 *dir, const filename3 name, nfsContext *con);

/* Copy an NFS file to a stream. Returns number of bytes written, or -1 on
   failure. */
int64 nfsRead (nfsObject *f, FILE *stream, nfsContext *con);

/* Copy a stream to an NFS file. Returns number of bytes written, or -1 on
   failure. */
int64 nfsWrite (nfsObject *f, FILE *stream, nfsContext *con);

/* Get contents of a symbolic link. Returns NULL on failure. When you are
   done with it, pass the nfspath to free(). */
nfspath nfsReadlink (nfs_fh3 *file, nfsContext *con);

/* Create a hard link. Returns -1 on failure, 0 on success. */
int nfsLink (const nfs_fh3 *dir, const filename3 from, const nfs_fh3 *to, nfsContext *con);

/* Create a symbolic link. Returns -1 on failure, 0 on success. */
int nfsSymlink (const nfs_fh3 *dir, const filename3 from, const filename3 to, nfsContext *con);

/* Rename an NFS object. Returns -1 on failure, 0 on success. */
int nfsRename (const nfs_fh3 *sdir, const filename3 sname, const nfs_fh3 *ddir, const filename3 dname, nfsContext *con);

/* Get NFS3 static filesystem info. When you are done with it, pass the
   FSINFO3resok to free(). */
FSINFO3resok *nfsFsinfo (const nfs_fh3 *fh, nfsContext *con);

/* Get NFS3 dynamic filesystem info. When you are done with it, pass the
   FSSTAT3resok to free(). */
FSSTAT3resok *nfsFsstat (const nfs_fh3 *fh, nfsContext *con);


/* Create NFS context. */
nfsContext *nfsNewContext (const nfs_fh3 *root, CLIENT *client, const u_long version);

/* Get path to cwd. When you are done with it, pass the filename3 to free(). */
filename3 nfsGetPath (const nfsContext *con);

#define	nfsGetVersion(c)	((c)->version)
#define	nfsGetUser(c)		((c)->authUid)
#define	nfsGetGroup(c)		((c)->authGid)
#define	nfsGetAtime(c)		((c)->useAtime ? ((c)->atime) : time (0))
#define	nfsGetMtime(c)		((c)->useMtime ? ((c)->mtime) : time (0))
#define	nfsGetUmask(c)		((int) (((c)->umask)&0777))
#define	nfsGetCwd(c)		((c)->cwd->fh)
#define	nfsGetRoot(c)		((c)->root->fh)

/* Change user credential. */
void nfsSetUser (uid_t uid, nfsContext *con);

/* Change group credential. */
void nfsSetGroup (gid_t gid, nfsContext *con);

/* Change context atime. */
void nfsSetAtime (const time_t *time, nfsContext *con);

/* Change context mtime. */
void nfsSetMtime (const time_t *time, nfsContext *con);

/* Change umask. */
#define	nfsSetUmask(u,c)	((void)((c)->umask=(u)))

/* Change remote directory. Returns 0 on success, -1 on failure. Leaves old
   cwd intact on failure. */
int nfsChdir (const filename3 name, nfsContext *con);

/* Change remote root directory. Returns 0 on success, -1 on failure. Leaves
   old root directory intact on failure. */
int nfsChroot (const filename3 name, nfsContext *con);


#endif	/* !defined(_NFSPROCS_H) */

